#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gitlabracadabra.objects.group import GitLabracadabraGroup
from gitlabracadabra.objects.project import GitLabracadabraProject
from gitlabracadabra.parser import GitlabracadabraParser
from gitlabracadabra.tests import patch_open
from gitlabracadabra.tests.case import TestCase


class TestParser(TestCase):
    def test_from_yaml_file(self):
        contents_map = {
            'gitlabracadabra.yml': """
                message:
                  hello: world
                """,
        }
        with patch_open(contents_map):
            p = GitlabracadabraParser.from_yaml_file('gitlabracadabra.yml')
            d = {k: v for k, v in p._items()}
            self.assertEqual(d, {'message': {'hello': 'world'}})

    def test_include_str(self):
        contents_map = {
            'gitlabracadabra.yml': """
                include:
                  - included.yml
                """,
            'included.yml': """
                message:
                  hello: world
                """,
        }
        with patch_open(contents_map):
            p = GitlabracadabraParser.from_yaml_file('gitlabracadabra.yml')
            d = {k: v for k, v in p._items()}
            self.assertEqual(d, {'message': {'hello': 'world'}})

    def test_include_local(self):
        contents_map = {
            'gitlabracadabra.yml': """
                include:
                  - local: included.yml
                """,
            'included.yml': """
                message:
                  hello: world
                """,
        }
        with patch_open(contents_map):
            p = GitlabracadabraParser.from_yaml_file('gitlabracadabra.yml')
            d = {k: v for k, v in p._items()}
            self.assertEqual(d, {'message': {'hello': 'world'}})

    def test_include_recursion(self):
        contents_map = {
            'gitlabracadabra.yml': """
                include:
                  - included.yml
                """,
            'included.yml': """
                include:
                  - gitlabracadabra.yml
                """,
        }
        with patch_open(contents_map):
            with self.assertRaises(ValueError) as cm:
                p = GitlabracadabraParser.from_yaml_file('gitlabracadabra.yml')
                {k: v for k, v in p._items()}

        self.assertEqual(str(cm.exception), 'gitlabracadabra.yml: nesting too deep in `include`')

    def test_include_forbidden_relative_path(self):
        contents_map = {
            'gitlabracadabra.yml': """
                include:
                  - ../included.yml
                """,
        }
        with patch_open(contents_map):
            with self.assertRaises(ValueError) as cm:
                p = GitlabracadabraParser.from_yaml_file('gitlabracadabra.yml')
                {k: v for k, v in p._items()}

        self.assertEqual(str(cm.exception), 'gitlabracadabra.yml: forbidden path for `include`: ../included.yml')

    def test_include_forbidden_absolute_path(self):
        contents_map = {
            'gitlabracadabra.yml': """
                include:
                  - /included.yml
                """,
        }
        with patch_open(contents_map):
            with self.assertRaises(ValueError) as cm:
                p = GitlabracadabraParser.from_yaml_file('gitlabracadabra.yml')
                {k: v for k, v in p._items()}

        self.assertEqual(str(cm.exception), 'gitlabracadabra.yml: forbidden path for `include`: /included.yml')

    def test_extend(self):
        p = GitlabracadabraParser.from_yaml('-', """
            .hidden-key:
              param1: value1
              param2: value2
            shown-key:
              extends: .hidden-key
              param3: value3
            another-key:
              param4: value4
            """)
        d = {k: v for k, v in p._items()}

        self.assertNotIn('.hidden-key', d)
        self.assertIn('shown-key', d)
        self.assertIn('another-key', d)
        self.assertEqual(d['another-key'], {'param4': 'value4'})

    def test_extend_not_found(self):
        p = GitlabracadabraParser.from_yaml('-', """
            key2:
              extends:
                - key1
            """)
        with self.assertRaises(ValueError) as cm:
            {k: v for k, v in p._items()}

        self.assertEqual(str(cm.exception), '- (`key1` from `key2`): key1 not found')

    def test_extend_recursion(self):
        p = GitlabracadabraParser.from_yaml('-', """
            key1:
              extends: key3
              foo: bar
            key2:
              extends: key1
              foo2: bar2
            key3:
              extends: key2
              foo3: bar3
            """)
        with self.assertRaises(ValueError) as cm:
            {k: v for k, v in p._items()}

        self.assertEqual(str(cm.exception), '- (key1): nesting too deep in `extends`')

    def test_extend_unknown_merge_strategy(self):
        p = GitlabracadabraParser.from_yaml('-', """
            key1:
              hello: world
            key2:
              extends:
                - key1: keep-the-best
            """)
        with self.assertRaises(ValueError) as cm:
            {k: v for k, v in p._items()}

        self.assertEqual(str(cm.exception), '- (`key1` from `key2`): Unknown merge strategy `keep-the-best`')

    def test_include_and_extend(self):
        contents_map = {
            'gitlabracadabra.yml': """
                group1/project1:
                  name: foo
                  extends: .e1
                  members:
                   baz: owner

                .e1:
                  branches:
                    - master
                    - develop
                  members:
                    bar: maintainer

                include:
                  - local: included.yml
                """,
            'included.yml': """
                .e1:
                  branches:
                    - master
                    - prod
                    - staging
                  members:
                    foo: developer
                """,
        }
        with patch_open(contents_map):
            p = GitlabracadabraParser.from_yaml_file('gitlabracadabra.yml')
            d = {k: v for k, v in p._items()}

        self.assertNotIn('.e1', d)
        self.assertNotIn('include', d)
        self.assertEqual(d['group1/project1'], {
            'name': 'foo',
            'members': {'foo': 'developer', 'bar': 'maintainer', 'baz': 'owner'},
            'branches': ['master', 'develop'],
        })

    def test_extend_list(self):
        contents_map = {
            'gitlabracadabra.yml': """
                group1/project1:
                  name: foo
                  extends:
                    - .e1
                    - .e2: replace
                    - .e3
                  members:
                   foo: owner

                .e1:
                  branches:
                    - master1
                    - develop1
                  members:
                    bar: maintainer

                .e2:
                  branches:
                    - master2
                    - develop2
                  members:
                    baz: developer

                .e3:
                  branches:
                    - master3
                    - develop3
                  members:
                    bar: developer # override .e1 maintainer
                """,
        }
        with patch_open(contents_map):
            p = GitlabracadabraParser.from_yaml_file('gitlabracadabra.yml')
            d = {k: v for k, v in p._items()}

        self.assertNotIn('.e1', d)
        self.assertNotIn('include', d)
        self.assertEqual(d['group1/project1'], {
            'name': 'foo',
            'members': {'foo': 'owner', 'bar': 'developer'},
            'branches': ['master3', 'develop3'],
        })

    def test_extend_list_aggregate(self):
        contents_map = {
            'gitlabracadabra.yml': """
                group1/project1:
                  name: foo
                  extends:
                    - .e1
                    - .e2: aggregate
                    - .e3
                  members:
                   foo: owner

                .e1:
                  branches:
                    - master1
                    - develop1
                  members:
                    bar: maintainer

                .e2:
                  branches:
                    - master2
                    - develop2
                  members:
                    baz: developer

                .e3:
                  branches:
                    - master3
                    - develop3
                  members:
                    bar: developer # override .e1 maintainer
                """,
        }
        with patch_open(contents_map):
            p = GitlabracadabraParser.from_yaml_file('gitlabracadabra.yml')
            d = {k: v for k, v in p._items()}

        self.assertNotIn('.e1', d)
        self.assertNotIn('include', d)
        self.assertEqual(d['group1/project1'], {
            'name': 'foo',
            'members': {'foo': 'owner', 'bar': 'developer', 'baz': 'developer'},
            'branches': ['master2', 'develop2', 'master3', 'develop3'],
        })

    def test_objects(self):
        p = GitlabracadabraParser.from_yaml('-', """
            .project-template:
              wiki_enabled: true
              issues_enabled: true
            group1/:
              description: My group
            group1/project1:
              extends: .project-template
              wiki_enabled: false
              description: My project
              buggy_param: Oh no!
            """)
        d = p.objects()

        self.assertNotIn('.project-template', d)

        self.assertIn('group1', d)
        group1 = d['group1']
        self.assertIsInstance(group1, GitLabracadabraGroup)
        self.assertEqual(group1._errors, [])
        self.assertEqual(group1._content, {'description': 'My group'})

        self.assertIn('group1/project1', d)
        project1 = d['group1/project1']
        self.assertIsInstance(project1, GitLabracadabraProject)
        self.assertEqual(len(project1._errors), 1)
        self.assertEqual(project1._errors[0].message, 'Additional properties are not allowed '
                                                      "('buggy_param' was unexpected)")
        self.assertEqual(project1._content, {'buggy_param': 'Oh no!',
                                             'description': 'My project',
                                             'issues_enabled': True,
                                             'wiki_enabled': False})



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2021 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


# =============================================================================
# Media Types (i.e. MIME types)
# =============================================================================

OCI_IMAGE_MANIFEST = 'application/vnd.oci.image.manifest.v1+json'
OCI_IMAGE_INDEX = 'application/vnd.oci.image.index.v1+json'

DOCKER_MANIFEST_SCHEMA2 = 'application/vnd.docker.distribution.manifest.v2+json'
DOCKER_MANIFEST_SCHEMA1_SIGNED = 'application/vnd.docker.distribution.manifest.v1+prettyjws'
DOCKER_MANIFEST_SCHEMA1 = 'application/vnd.docker.distribution.manifest.v1+json'
DOCKER_MANIFEST_SCHEMA2_LIST = 'application/vnd.docker.distribution.manifest.list.v2+json'

MANIFEST = (
    OCI_IMAGE_MANIFEST,
    DOCKER_MANIFEST_SCHEMA2,
    DOCKER_MANIFEST_SCHEMA1_SIGNED,
    DOCKER_MANIFEST_SCHEMA1,
    DOCKER_MANIFEST_SCHEMA2_LIST,
    OCI_IMAGE_INDEX,
)

MANIFEST_LIST = (
    DOCKER_MANIFEST_SCHEMA2_LIST,
    # OCI_IMAGE_INDEX,
)

# =============================================================================
# Domains
# =============================================================================

DOCKER_HOSTNAME = 'docker.io'
DOCKER_REGISTRY = 'registry-1.docker.io'

# =============================================================================
# Headers
# =============================================================================

DIGEST_HEADER = 'Docker-Content-Digest'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from unittest import skipIf

from gitlab import __version__ as gitlab_version

from gitlabracadabra.objects.project import GitLabracadabraProject
from gitlabracadabra.objects.user import GitLabracadabraUser
from gitlabracadabra.tests import my_vcr, patch
from gitlabracadabra.tests.case import TestCaseWithManager


class TestProject(TestCaseWithManager):
    @my_vcr.use_cassette
    def test_no_create(self, cass):
        obj = GitLabracadabraProject('memory', 'test/no_create_object', {})
        obj.process()
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_create(self, cass):
        obj = GitLabracadabraProject('memory', 'test/create_object', {'create_object': True})
        obj.process()
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_delete(self, cass):
        obj = GitLabracadabraProject('memory', 'test/delete_me', {'delete_object': True})
        obj.process()
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_exists(self, cass):
        obj = GitLabracadabraProject('memory', 'test/exists', {})
        obj.process()
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_simple_parameters(self, cass):
        obj = GitLabracadabraProject('memory', 'test/project_simple_parameters', {
            'name': 'project-with-simple-parameters',
            'description': 'Project with simple parameters',
            'issues_enabled': False,
            'merge_requests_enabled': False,
            'jobs_enabled': False,
            'wiki_enabled': False,
            'snippets_enabled': False,
            'resolve_outdated_diff_discussions': True,
            'container_registry_enabled': False,
            'shared_runners_enabled': False,
            'visibility': 'public',
            # 'import_url': 'http://example.com/foo.git',  # FIXME
            # 'public_builds': False,  # FIXME
            'only_allow_merge_if_pipeline_succeeds': True,
            'only_allow_merge_if_all_discussions_are_resolved': True,
            'merge_method': 'ff',
            # 'remove_source_branch_after_merge': False, # TODO
            'lfs_enabled': False,
            'request_access_enabled': True,
            'tag_list': ['foo', 'bar'],
            # 'avatar': 'http://example.com/foo.png',  # FIXME
            'printing_merge_request_link_enabled': False,
            'ci_config_path': 'debian/gitlab-ci.yml',
            # 'repository_storage': 'foo',  # EE
            # 'approvals_before_merge': False,  # EE
            # 'external_authorization_classification_label': 'foo',  # EE
            # 'mirror': False,  # EE
            # 'mirror_user_id': 2,  # EE
            # 'mirror_trigger_builds': False,  # EE
            # 'only_mirror_protected_branches': True,  # EE
            # 'mirror_overwrites_diverged_branches': True,  # EE
            # 'packages_enabled': True,  # EE
        })
        obj.process()
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_simple_parameters2(self, cass):
        obj = GitLabracadabraProject('memory', 'test/project_simple_parameters2', {
            'issues_access_level': 'private',
            'repository_access_level': 'private',
            'merge_requests_access_level': 'private',
            'builds_access_level': 'private',
            'wiki_access_level': 'private',
            'snippets_access_level': 'private',
            'build_git_strategy': 'clone',
            'build_timeout': 7200,
            'auto_cancel_pending_pipelines': 'disabled',
            'build_coverage_regex': '^TOTAL\\s+\\d+\\s+\\d+\\s+(\\d+\\%)$',
            'ci_default_git_depth': 100,
            'auto_devops_enabled': False,
            'auto_devops_deploy_strategy': 'timed_incremental',
        })
        obj.process()
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_default_branch_exists(self, cass):
        obj = GitLabracadabraProject('memory', 'test/project_default_branch', {
            'default_branch': 'exists',
        })
        obj.process()
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_default_branch_not_exists(self, cass):
        obj = GitLabracadabraProject('memory', 'test/project_default_branch', {
            'default_branch': 'not_exists',
        })
        with patch('gitlabracadabra.objects.object.logger', autospec=True) as logger:
            obj.process()
            self.assertTrue(cass.all_played)
            logger.error.assert_called_once_with('[%s] Unable to change param %s (%s -> %s): %s',
                                                 'test/project_default_branch',
                                                 'default_branch',
                                                 'exists',
                                                 'not_exists',
                                                 {'base': [
                                                  "Could not change HEAD: branch 'not_exists' does not exist"]})

    @my_vcr.use_cassette
    def test_branches(self, cass):
        obj = GitLabracadabraProject('memory', 'test/project_branches', {
            'branches': ['a', 'b', 'c'],

        })
        obj.process()
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_groups(self, cass):
        obj = GitLabracadabraProject('memory', 'test/project_groups', {
            'groups': {'test2': 'developer'},
        })
        obj.process()
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_groups_change_access_level(self, cass):
        obj = GitLabracadabraProject('memory', 'test/project_groups', {
            'groups': {'test2': 'maintainer'},
        })
        obj.process()
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_groups_not_found(self, cass):
        obj = GitLabracadabraProject('memory', 'test/project_groups', {
            'groups': {'group_not_found': 'maintainer'},
            'unknown_groups': 'ignore',
        })
        with patch('gitlabracadabra.mixins.groups.logger', autospec=True) as logger:
            obj.process()
            self.assertTrue(cass.all_played)
            logger.warning.assert_called_once_with('[%s] Group not found %s',
                                                   'test/project_groups',
                                                   'group_not_found')

    @my_vcr.use_cassette
    def test_groups_delete_unknown(self, cass):
        obj = GitLabracadabraProject('memory', 'test/project_groups', {
            'groups': {},
            'unknown_groups': 'delete',
        })
        obj.process()
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_members(self, cass):
        # Clean up
        GitLabracadabraUser._USERS_USERNAME2ID = {}
        GitLabracadabraUser._USERS_ID2USERNAME = {}
        obj = GitLabracadabraProject('memory', 'test/project_members', {
            'members': {'some_member': 'developer'},
        })
        obj.process()
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_members_change_access_level(self, cass):
        obj = GitLabracadabraProject('memory', 'test/project_members', {
            'members': {'some_member': 'maintainer'},
        })
        obj.process()
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_members_not_found(self, cass):
        # Clean up
        GitLabracadabraUser._USERS_USERNAME2ID = {}
        GitLabracadabraUser._USERS_ID2USERNAME = {}
        obj = GitLabracadabraProject('memory', 'test/project_members', {
            'members': {'member_not_found': 'maintainer'},
            'unknown_members': 'ignore',
        })
        with patch('gitlabracadabra.mixins.members.logger', autospec=True) as logger:
            obj.process()
            self.assertTrue(cass.all_played)
            logger.warning.assert_called_once_with('[%s] User not found %s',
                                                   'test/project_members', 'member_not_found')

    @my_vcr.use_cassette
    def test_members_delete_unknown(self, cass):
        obj = GitLabracadabraProject('memory', 'test/project_members', {
            'members': {},
            'unknown_members': 'delete',
        })
        obj.process()
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_protected_branches_wildcard(self, cass):
        obj = GitLabracadabraProject('memory', 'test/protected_branches', {
            'protected_branches': {'release/*': {'push_access_level': 'noone', 'merge_access_level': 'maintainer'}},
        })
        with patch('gitlabracadabra.objects.project.logger', autospec=True) as logger:
            obj.process()
            self.assertTrue(cass.all_played)
            logger.warning.assert_called_once_with('[%s] NOT Deleting unknown protected branch: %s'
                                                   ' (unknown_protected_branches=%s)',
                                                   'test/protected_branches', 'master', 'warn')

    @my_vcr.use_cassette
    def test_protected_branches_delete(self, cass):
        obj = GitLabracadabraProject('memory', 'test/protected_branches', {
            'protected_branches': {},
            'unknown_protected_branches': 'delete',
        })
        with patch('gitlabracadabra.objects.project.logger', autospec=True) as logger:
            obj.process()
            self.assertTrue(cass.all_played)
            logger.info.assert_called_once_with('[%s] Deleting unknown protected branch: %s',
                                                'test/protected_branches', 'master')

    @skipIf(gitlab_version in ['1.6.0'], 'python-gitlab without protected tags support')
    @my_vcr.use_cassette
    def test_protected_tags_wildcard(self, cass):
        obj = GitLabracadabraProject('memory', 'test/protected_tags', {
            'protected_tags': {'v*': 'maintainer'},
        })
        with patch('gitlabracadabra.objects.project.logger', autospec=True) as logger:
            obj.process()
            self.assertTrue(cass.all_played)
            logger.info.assert_called_once_with('[%s] Changing protected tag %s access level: %s -> %s',
                                                'test/protected_tags', 'v*',
                                                {}, {'name': 'v*', 'create_access_level': 40})

    @skipIf(gitlab_version in ['1.6.0'], 'python-gitlab without protected tags support')
    @my_vcr.use_cassette
    def test_protected_tags_change(self, cass):
        obj = GitLabracadabraProject('memory', 'test/protected_tags', {
            'protected_tags': {'v1.0': 'maintainer'},
        })
        with patch('gitlabracadabra.objects.project.logger', autospec=True) as logger:
            obj.process()
            self.assertTrue(cass.all_played)
            logger.info.assert_called_once_with('[%s] Changing protected tag %s access level: %s -> %s',
                                                'test/protected_tags', 'v1.0',
                                                {'name': 'v1.0', 'create_access_level': 30},
                                                {'name': 'v1.0', 'create_access_level': 40})

    @skipIf(gitlab_version in ['1.6.0'], 'python-gitlab without protected tags support')
    @my_vcr.use_cassette
    def test_protected_tags_delete(self, cass):
        obj = GitLabracadabraProject('memory', 'test/protected_tags', {
            'protected_tags': {},
            'unknown_protected_tags': 'delete',
        })
        with patch('gitlabracadabra.objects.project.logger', autospec=True) as logger:
            obj.process()
            self.assertTrue(cass.all_played)
            logger.info.assert_called_once_with('[%s] Deleting unknown protected tag: %s',
                                                'test/protected_tags', 'unknown')

    @my_vcr.use_cassette
    def test_archived(self, cass):
        obj = GitLabracadabraProject('memory', 'test/archived', {
            'archived': True,
        })
        with patch('gitlabracadabra.objects.project.logger', autospec=True) as logger:
            obj.process()
            self.assertTrue(cass.all_played)
            logger.info.assert_called_once_with('[%s] Changing param %s: %s -> %s',
                                                'test/archived', 'archived', False, True)

    @my_vcr.use_cassette
    def test_unarchived(self, cass):
        obj = GitLabracadabraProject('memory', 'test/unarchived', {
            'archived': False,
        })
        with patch('gitlabracadabra.objects.project.logger', autospec=True) as logger:
            obj.process()
            self.assertTrue(cass.all_played)
            logger.info.assert_called_once_with('[%s] Changing param %s: %s -> %s',
                                                'test/unarchived', 'archived', True, False)

    @my_vcr.use_cassette
    def test_variables(self, cass):
        obj = GitLabracadabraProject('memory', 'test/project_variables', {
            'variables': [
                {
                    'key': 'simple_var',
                    'value': 'simple_value',
                },
                {
                    'key': 'file_variable',
                    'value': 'BEGIN CERTIFICATE',
                    'variable_type': 'file',
                },
                {
                    'key': 'change_me',
                    'value': 'new_value',
                    'masked': False,
                    'protected': False,
                    'variable_type': 'file',
                },
            ],
            'unknown_variables': 'ignore',
        })
        obj.process()
        self.assertTrue(cass.all_played)

    @skipIf(gitlab_version in ['1.11.0', '1.13.0'], 'python-gitlab without proper project labels support')
    @my_vcr.use_cassette
    def test_labels(self, cass):
        obj = GitLabracadabraProject('memory', 'test/project_labels', {
            'labels': [
                {
                    'name': 'new_label',
                    'color': 'red',
                },
                {
                    'name': 'change_me',
                    'color': 'green',
                    'description': 'New description',
                },
            ],
            'unknown_labels': 'delete',
        })
        obj.process()
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_milestones(self, cass):
        obj = GitLabracadabraProject('memory', 'test/project_milestones', {
            'milestones': [
                {
                    'title': 'new_milestone',
                    'description': 'New milestone',
                    'due_date': '2023-01-23',
                    'start_date': '2022-01-23',
                    # 'state': 'active',  # FIXME
                },
                {
                    'title': 'existing_milestone',
                    'description': 'New description',
                    'due_date': '',
                    'start_date': '',
                    'state': 'closed',
                },
            ],
            'unknown_milestones': 'delete',
        })
        obj.process()
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_webhooks(self, cass):
        obj = GitLabracadabraProject('memory', 'test/test_webhooks', {
            'webhooks': [
                {
                    'url': 'http://example.com/create_me',
                    'push_events': False,
                    'push_events_branch_filter': 'master',
                    'issues_events': True,
                    'confidential_issues_events': True,
                    'merge_requests_events': True,
                    'tag_push_events': True,
                    'note_events': True,
                    'confidential_note_events': True,
                    'job_events': True,
                    'pipeline_events': True,
                    'wiki_page_events': True,
                    'enable_ssl_verification': False,
                    # 'repository_update_events': True,
                },
                {
                    'url': 'http://example.com/modify_me',
                    'push_events': False,
                    'push_events_branch_filter': 'master',
                    'issues_events': True,
                    'confidential_issues_events': True,
                    'merge_requests_events': True,
                    'tag_push_events': True,
                    'note_events': True,
                    'confidential_note_events': True,
                    'job_events': True,
                    'pipeline_events': True,
                    'wiki_page_events': True,
                    'enable_ssl_verification': False,
                    # 'repository_update_events': True,
                },
            ],
            'unknown_webhooks': 'delete',
        })
        obj.process()
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_container_expiration_policy(self, cass):
        obj = GitLabracadabraProject('memory', 'test/container_expiration_policy', {
            'container_expiration_policy': {
                'enabled': True,
                'cadence': '14d',
                'keep_n': 25,
                'name_regex_keep': '.*master|.*release|release-.*|master-.*',
                'older_than': '90d',
                'name_regex_delete': '.*',
            },
        })
        obj.process()
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_pipeline_schedules(self, cass):
        obj = GitLabracadabraProject('memory', 'test/test_pipeline_schedules', {
            'pipeline_schedules': [
                {
                    'description': 'create_me',
                    'ref': 'master',
                    'cron': '0 1 * * 5',
                    'cron_timezone': 'UTC',
                    'active': True,
                },
                {
                    'description': 'modify_me',
                    'ref': 'develop',
                    'cron': '0 3 * * 5',
                    'cron_timezone': 'Pacific Time (US & Canada)',
                    'active': False,
                },
            ],
            'unknown_pipeline_schedules': 'delete',
            'unknown_pipeline_schedule_variables': 'ignore',
        })
        obj.process()
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_pipeline_schedule_variables(self, cass):
        obj = GitLabracadabraProject('memory', 'test/test_pipeline_schedule_variables', {
            'pipeline_schedules': [
                {
                    'description': 'some_schedule',
                    'ref': 'master',
                    'cron': '0 16 * * *',
                    'cron_timezone': 'UTC',
                    'active': False,
                    'variables': [
                        {
                            'key': 'create_me',
                            'value': 'some_value',
                            'variable_type': 'file',
                        },
                        {
                            'key': 'modify_me',
                            'value': 'new_value',
                            'variable_type': 'file',
                        },
                    ],
                },
            ],
            'unknown_pipeline_schedules': 'delete',
            'unknown_pipeline_schedule_variables': 'delete',
        })
        obj.process()
        self.assertTrue(cass.all_played)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2021 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import List

from gitlabracadabra.containers.blob import Blob
from gitlabracadabra.containers.const import (
    DOCKER_MANIFEST_SCHEMA1,
    DOCKER_MANIFEST_SCHEMA1_SIGNED,
    DOCKER_MANIFEST_SCHEMA2,
    DOCKER_MANIFEST_SCHEMA2_LIST,
    OCI_IMAGE_INDEX,
    OCI_IMAGE_MANIFEST,
)
from gitlabracadabra.containers.manifest_base import ManifestBase


class Manifest(ManifestBase):
    """Retrieve Manifest or Manifest list."""

    def manifests(self) -> List[Manifest]:
        """Get manifests of the manifest list.

        Returns:
            A list of manifests.

        Raises:
            ValueError: Unsupported manifest list type.
        """
        if self.mime_type in {DOCKER_MANIFEST_SCHEMA2_LIST, OCI_IMAGE_INDEX}:
            return self._manifests_v2()
        raise ValueError('Unsupported manifest list type {0}'.format(self.mime_type))

    def tag_list(self) -> List[str]:
        """Get tags of the manifest.

        Returns:
            A list of tags (strings).

        Raises:
            ValueError: Expected list got something else.
        """
        response = self._registry.request(
            'get',
            '/v2/{0}/tags/list'.format(self.manifest_name),
            scopes={self.scope()},
        )
        tags = response.json().get('tags')
        if not isinstance(tags, list):
            raise ValueError('Expected list got {0}'.format(type(tags)))
        return tags

    def blobs(self) -> List[Blob]:
        """Get blobs of the manifest.

        Returns:
            A list of blobs.

        Raises:
            ValueError: Unsupported media type.
        """
        if self.mime_type in {DOCKER_MANIFEST_SCHEMA2, OCI_IMAGE_MANIFEST}:
            blobs = []
            for layer_json in self.json.get('layers'):
                blobs.append(Blob(
                    self.registry,
                    self.manifest_name,
                    layer_json['digest'],
                    size=layer_json['size'],
                    mime_type=layer_json['mediaType'],
                ))
            return blobs
        elif self.mime_type in {DOCKER_MANIFEST_SCHEMA1, DOCKER_MANIFEST_SCHEMA1_SIGNED}:
            blobs = []
            for fs_layer_json in self.json.get('fsLayers'):
                blobs.append(Blob(
                    self.registry,
                    self.manifest_name,
                    fs_layer_json['blobSum'],
                    mime_type='application/octet-stream',
                ))
            return blobs
        raise ValueError('Unsupported media type: {0}'.format(self.mime_type))

    def _manifests_v2(self) -> List[Manifest]:
        json = dict(self.json)
        if json['mediaType'] not in {DOCKER_MANIFEST_SCHEMA2_LIST, OCI_IMAGE_INDEX}:
            raise ValueError('Unexpected manifest list type {0}'.format(json['mediaType']))
        if json['schemaVersion'] != 2:
            raise ValueError('Unexpected manifest schema version {0}'.format(json['schemaVersion']))
        manifests = []
        for manifest_json in json['manifests']:
            manifest = Manifest(
                self.registry,
                self.manifest_name,
                digest=manifest_json['digest'],
                size=manifest_json['size'],
                mime_type=manifest_json['mediaType'],
                tag=self.tag,
            )
            manifest.platform = manifest_json['platform']
            manifests.append(manifest)
        return manifests



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2021 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from copy import deepcopy

from gitlabracadabra.gitlab.access_levels import access_level


logger = logging.getLogger(__name__)


class GroupsMixin(object):
    """Object with groups."""

    """_process_groups()

    Process the groups param.
    """
    def _process_groups(self, param_name, param_value, dry_run=False, skip_save=False):
        assert param_name == 'groups'  # noqa: S101
        assert not skip_save  # noqa: S101
        if not hasattr(self._obj, 'shared_with_groups'):
            logger.error('[%s] Unable to share with groups: %s',
                         self._name, 'share group with groups requires GitLab >= 13.1.0')
            return
        if not hasattr(self._obj, 'share'):
            # https://github.com/python-gitlab/python-gitlab/pull/1139
            logger.error('[%s] Unable to share with groups: %s',
                         self._name, 'share group with groups requires python-gitlab >= 2.5.0')
            return
        param_value = deepcopy(param_value)
        unknown_groups = self._content.get('unknown_groups', 'warn')
        # We first check for already shared groups
        for group in self._obj.shared_with_groups:
            if 'group_full_path' not in group:
                # Gitlab < 11.8
                # https://gitlab.com/gitlab-org/gitlab-ce/merge_requests/24052
                group['group_full_path'] = self.connection.group_cache.full_path_from_id(group['group_id'])
            self.connection.group_cache.map_group(group['group_id'], group['group_full_path'])
            if group['group_full_path'] in param_value.keys():
                target_access_level = access_level(param_value[group['group_full_path']])
                if group['group_access_level'] != target_access_level:
                    if dry_run:
                        logger.info('[%s] NOT Changing group %s access level: %s -> %s (dry-run)',
                                    self._name, group['group_full_path'],
                                    group['group_access_level'], target_access_level)
                    else:
                        logger.info('[%s] Changing group %s access level: %s -> %s',
                                    self._name, group['group_full_path'],
                                    group['group_access_level'], target_access_level)
                        self._obj.unshare(group['group_id'])
                        self._obj.share(group['group_id'], target_access_level)
                param_value.pop(group['group_full_path'])
            else:
                if unknown_groups in ['delete', 'remove']:
                    if dry_run:
                        logger.info('[%s] NOT Unsharing from unknown group: %s (dry-run)',
                                    self._name, group['group_full_path'])
                    else:
                        logger.info('[%s] Unsharing from unknown group: %s',
                                    self._name, group['group_full_path'])
                        self._obj.unshare(group['group_id'])
                elif unknown_groups not in ['ignore', 'skip']:
                    logger.warning('[%s] NOT Unsharing from unknown group: %s (unknown_groups=%s)',
                                   self._name, group['group_full_path'], unknown_groups)
        # Remaining groups
        for group_full_path, target_group_access in sorted(param_value.items()):
            group_id = self.connection.group_cache.id_from_full_path(group_full_path)
            if group_id is None:
                logger.warning('[%s] Group not found %s',
                               self._name, group_full_path)
                continue
            target_access_level = access_level(target_group_access)
            if dry_run:
                logger.info('[%s] NOT Sharing group %s: %s -> %s (dry-run)',
                            self._name, group_full_path,
                            0, target_access_level)
            else:
                logger.info('[%s] Sharing group %s: %s -> %s',
                            self._name, group_full_path,
                            0, target_access_level)
                self._obj.share(group_id, target_access_level)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from unittest.mock import call, patch
from urllib.parse import quote

from pygit2 import init_repository

from gitlabracadabra.objects.project import GitLabracadabraProject
from gitlabracadabra.tests import my_vcr
from gitlabracadabra.tests.case import TestCaseWithManager


class TestProjectMirror(TestCaseWithManager):
    @my_vcr.use_cassette
    def test_mirrors_pull(self, cass):
        testrepo_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'fixtures', 'testrepo.git')
        testrepo_url = 'file://{0}'.format(testrepo_dir)
        dest_dir = str(self._temp_dir / 'dest.git')
        dest_url = 'file://{0}'.format(dest_dir)
        dest_repo = init_repository(dest_dir, bare=True)
        repo_cache_dir = str(self._temp_dir / 'cache' / 'gitlabracadabra' / quote(dest_url, safe=''))
        with patch('gitlabracadabra.mixins.mirrors.logger', autospec=True) as logger:
            with patch.object(GitLabracadabraProject, 'web_url') as web_url_mock:
                project = GitLabracadabraProject('memory', 'test/project_pull_mirror', {
                    'create_object': True,
                    'mirrors': [
                        {
                            'url': testrepo_url,
                            'direction': 'pull',
                        },
                    ],
                })

                web_url_mock.return_value = dest_url
                project.process()
                logger.debug.assert_called_once_with(
                    '[%s] Creating cache repository in %s',
                    'test/project_pull_mirror',
                    repo_cache_dir,
                )
                logger.info.assert_has_calls(
                    [
                        call(
                            '[%s] %s Pushing %s %s to %s: %s -> %s',
                            'test/project_pull_mirror',
                            testrepo_url,
                            'branch',
                            'hello',
                            'hello',
                            None,
                            '8d1fd4e584faf465d96e2f9b3cbd5000721469b3',
                        ),
                        call(
                            '[%s] %s Pushing %s %s to %s: %s -> %s',
                            'test/project_pull_mirror',
                            testrepo_url,
                            'branch',
                            'master',
                            'master',
                            None,
                            '5e8dfc288cf87620e22e67b6db671dc8a596e2f9',
                        ),
                        call(
                            '[%s] %s Pushing %s %s to %s: %s -> %s',
                            'test/project_pull_mirror',
                            testrepo_url,
                            'tag',
                            'tag1',
                            'tag1',
                            None,
                            '8d1fd4e584faf465d96e2f9b3cbd5000721469b3',
                        ),
                        call(
                            '[%s] %s Pushing %s %s to %s: %s -> %s',
                            'test/project_pull_mirror',
                            testrepo_url,
                            'tag',
                            'tag2',
                            'tag2',
                            None,
                            '5e8dfc288cf87620e22e67b6db671dc8a596e2f9',
                        ),
                    ],
                    any_order=True,
                )
        self.assertEqual(
            list(dest_repo.references),
            ['refs/heads/hello', 'refs/heads/master', 'refs/tags/tag1', 'refs/tags/tag2'],
        )
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_mirrors_pull_mappings(self, cass):
        testrepo_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'fixtures', 'testrepo.git')
        testrepo_url = 'file://{0}'.format(testrepo_dir)
        dest_dir = str(self._temp_dir / 'dest2.git')
        dest_url = 'file://{0}'.format(dest_dir)
        dest_repo = init_repository(dest_dir, bare=True)
        repo_cache_dir = str(self._temp_dir / 'cache' / 'gitlabracadabra' / quote(dest_url, safe=''))
        with patch('gitlabracadabra.mixins.mirrors.logger', autospec=True) as logger:
            with patch.object(GitLabracadabraProject, 'web_url') as web_url_mock:
                project = GitLabracadabraProject('memory', 'test/project_pull_mirror', {
                    'create_object': True,
                    'mirrors': [
                        {
                            'url': testrepo_url,
                            'direction': 'pull',
                            'branches': [
                                {
                                    'from': 'hello',
                                    'to': 'world',
                                },
                                {
                                    'from': '/(.*)/',
                                    'to': r'upstream/\1',
                                },
                            ],
                            'tags': [
                                {
                                    'from': '/(.*1)/',
                                    'to': r'upstream-\1',
                                },
                            ],
                        },
                    ],
                })

                web_url_mock.return_value = dest_url
                project.process()
                logger.debug.assert_called_once_with(
                    '[%s] Creating cache repository in %s',
                    'test/project_pull_mirror',
                    repo_cache_dir,
                )
                logger.info.assert_has_calls(
                    [
                        call(
                            '[%s] %s Pushing %s %s to %s: %s -> %s',
                            'test/project_pull_mirror',
                            testrepo_url,
                            'branch',
                            'hello',
                            'world',
                            None,
                            '8d1fd4e584faf465d96e2f9b3cbd5000721469b3',
                        ),
                        call(
                            '[%s] %s Pushing %s %s to %s: %s -> %s',
                            'test/project_pull_mirror',
                            testrepo_url,
                            'branch',
                            'master',
                            'upstream/master',
                            None,
                            '5e8dfc288cf87620e22e67b6db671dc8a596e2f9',
                        ),
                        call(
                            '[%s] %s Pushing %s %s to %s: %s -> %s',
                            'test/project_pull_mirror',
                            testrepo_url,
                            'tag',
                            'tag1',
                            'upstream-tag1',
                            None,
                            '8d1fd4e584faf465d96e2f9b3cbd5000721469b3',
                        ),
                    ],
                    any_order=True,
                )
        self.assertEqual(
            list(dest_repo.references),
            ['refs/heads/upstream/master', 'refs/heads/world', 'refs/tags/upstream-tag1'],
        )
        self.assertTrue(cass.all_played)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from unittest import skipIf

from gitlab import __version__ as gitlab_version

from gitlabracadabra.objects.project import GitLabracadabraProject
from gitlabracadabra.tests import my_vcr
from gitlabracadabra.tests.case import TestCaseWithManager


class TestProjectBoards(TestCaseWithManager):
    @skipIf(gitlab_version in ['1.6.0'], 'python-gitlab without board update support')
    @my_vcr.use_cassette
    def test_boards(self, cass):
        obj = GitLabracadabraProject('memory', 'test/test_boards', {
            'boards': [
                {
                    'name': 'create_me',
                    'hide_backlog_list': True,
                    'hide_closed_list': True,
                    'lists': [
                        {
                            'label': 'move_me_first',
                        },
                        {
                            'label': 'keep_me',
                        },
                        {
                            'label': 'add_me',
                        },
                        {
                            'label': 'i_dont_exists',
                        },
                    ],
                },
                {
                    'name': 'modify_me',
                    'hide_backlog_list': True,
                    'hide_closed_list': True,
                    'lists': [
                        {
                            'label': 'move_me_first',
                        },
                        {
                            'label': 'keep_me',
                        },
                        {
                            'label': 'add_me',
                        },
                        {
                            'label': 'i_dont_exists',
                        },
                    ],
                },
                {
                    'name': 'do_no_modify_my_lists',
                },
            ],
            'unknown_boards': 'delete',
            'unknown_board_lists': 'delete',
        })
        self.assertEqual(obj.errors(), [])
        obj.process()
        self.assertTrue(cass.all_played)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2021 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from os.path import isfile

from gitlabracadabra.containers.blob import Blob
from gitlabracadabra.containers.registry import Registry
from gitlabracadabra.tests import my_vcr
from gitlabracadabra.tests.case import TestCase


MANIFEST_NAME = 'debian'
DIGEST = 'sha256:12345'
SIZE = 42


class TestBlob(TestCase):
    def test_equal(self):  # noqa: WPS210
        registry1 = Registry('localhost:5000')
        registry2 = Registry('localhost:5000')
        blob1 = Blob(registry1, MANIFEST_NAME, DIGEST)
        blob2 = Blob(registry1, MANIFEST_NAME, DIGEST)
        blob3 = Blob(registry2, MANIFEST_NAME, DIGEST)
        blob4 = Blob(registry1, MANIFEST_NAME, DIGEST, size=SIZE)
        self.assertEqual(blob1, blob2)
        self.assertNotEqual(blob1, blob3)
        self.assertNotEqual(blob1, blob4)

    @my_vcr.use_cassette
    def test_open(self, cass):
        registry = Registry('docker.io')
        blob = Blob(
            registry,
            'library/debian',
            'sha256:5890f8ba95f680c87fcf89e51190098641b4f646102ce7ca906e7f83c84874dc',
        )
        self.assertIsInstance(blob, Blob)
        self.assertFalse(isfile(blob.cache_path))
        with blob as opened_blob:
            blob_content = opened_blob.read().decode('utf-8')
            self.assertTrue(blob_content.startswith('{"architecture":"amd64","config":{'), blob_content)
        self.assertTrue(isfile(blob.cache_path))
        self.assertTrue(cass.all_played)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2021 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from types import MappingProxyType
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from typing import Optional


LEVELS = MappingProxyType({
    'noone': 0,
    'guest': 10,
    'reporter': 20,
    'developer': 30,
    'maintainer': 40,
    'owner': 50,
})


def access_level(level: str) -> Optional[int]:
    """Convert access level to int.

    Args:
        level: Access level as str.

    Returns:
        Access level as int, or None.
    """
    return LEVELS.get(level, None)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2021 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from collections import namedtuple
from typing import Any
from unittest.mock import MagicMock

from gitlabracadabra.matchers import Matcher
from gitlabracadabra.tests.case import TestCaseWithManager


TestData = namedtuple('TestData', ['patterns', 'as_callable', 'result', 'called'])

INPUT_DATA = ('item', 'item_suffix', 'prefix_item', 'another')

TEST_DATA = (
    # String patterns
    TestData('item', as_callable=False, result=['item'], called=None),
    TestData('item', as_callable=True, result=['item'], called=False),
    TestData(['item', 'extra'], as_callable=False, result=['item'], called=None),
    TestData(['item', 'extra'], as_callable=True, result=['item', 'extra'], called=False),
    # Regex patterns
    TestData('/item/', as_callable=False, result=['item'], called=None),
    TestData('/item/', as_callable=True, result=['item'], called=True),
    TestData(['/item/', '/extra/'], as_callable=False, result=['item'], called=None),
    TestData(['/item/', '/extra/'], as_callable=True, result=['item'], called=True),
    # Mixed patterns
    TestData(['another', '/item.*/'], as_callable=False, result=['item', 'item_suffix', 'another'], called=None),
    TestData(['another', '/item.*/'], as_callable=True, result=['item', 'item_suffix', 'another'], called=True),
    TestData(['another', '/.*item/'], as_callable=False, result=['item', 'prefix_item', 'another'], called=None),
    TestData(['another', '/.*item/'], as_callable=True, result=['item', 'prefix_item', 'another'], called=True),
    # Flags
    TestData('/Item/', as_callable=False, result=[], called=None),
    TestData('/Item/i', as_callable=False, result=['item'], called=None),
)


class TestMatcher(TestCaseWithManager):
    """Test Matcher class."""

    def test_match(self):
        """Test Matcher.match method."""
        for test_data in TEST_DATA:
            with self.subTest(patterns=test_data.patterns, as_callable=test_data.as_callable):
                if test_data.as_callable:
                    input_data = MagicMock()
                    input_data.return_value = list(INPUT_DATA)
                else:
                    input_data = list(INPUT_DATA)
                matches = Matcher(test_data.patterns).match(input_data)
                self.assertEqual(len(matches), len(test_data.result))
                for index, match in enumerate(matches):
                    self.assertEqual(match[0], test_data.result[index])
                self._assert_call(input_data, test_data.as_callable, test_data.called)

    def _assert_call(self, input_data: Any, as_callable: bool, called: bool) -> None:
        if as_callable and called:
            input_data.assert_called_once_with()
        elif as_callable:
            self.assertEqual(len(input_data.mock_calls), 0)
        else:
            self.assertIsNone(called)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from unittest import skipIf

from gitlab import __version__ as gitlab_version

from gitlabracadabra.objects.group import GitLabracadabraGroup
from gitlabracadabra.objects.user import GitLabracadabraUser
from gitlabracadabra.tests import my_vcr, patch
from gitlabracadabra.tests.case import TestCaseWithManager


class TestGroup(TestCaseWithManager):
    @my_vcr.use_cassette
    def test_no_create(self, cass):
        obj = GitLabracadabraGroup('memory', 'test/no_create_group', {})
        obj.process()
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_delete(self, cass):
        obj = GitLabracadabraGroup('memory', 'delete_this_group', {'delete_object': True})
        obj.process()
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_create(self, cass):
        obj = GitLabracadabraGroup('memory', 'test/create_group', {'create_object': True})
        obj.process()
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_missing_parent(self, cass):
        obj = GitLabracadabraGroup('memory', 'test/missing_parent/subgroup', {'create_object': True})
        with patch('gitlabracadabra.objects.object.logger', autospec=True) as logger:
            obj.process()
            self.assertTrue(cass.all_played)
            logger.error.assert_called_once_with('[%s] NOT Creating %s (%s)',
                                                 'test/missing_parent/subgroup',
                                                 'group',
                                                 'parent namespace not found')

    @my_vcr.use_cassette
    def test_exists(self, cass):
        obj = GitLabracadabraGroup('memory', 'test/group_exists', {})
        obj.process()
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_simple_parameters(self, cass):
        obj = GitLabracadabraGroup('memory', 'test/group_simple_parameters', {
            'name': 'group-with-simple-parameters',
            'description': 'Group with simple parameters',
            # 'membership_lock': true,  # EE-only
            # 'share_with_group_lock': true,  # EE-only
            'visibility': 'public',
            # 'file_template_project_id': 12,  # EE-only
            'lfs_enabled': False,
            'request_access_enabled': True,
            # 'shared_runners_minutes_limit': 42,  # EE, admin-only
            # 'extra_shared_runners_minutes_limit': 42,  # EE, admin-only
        })
        obj.process()
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_members(self, cass):
        obj = GitLabracadabraGroup('memory', 'test/group_members', {
            'members': {'some_member': 'developer'},
        })
        obj.process()
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_members_change_access_level(self, cass):
        obj = GitLabracadabraGroup('memory', 'test/group_members', {
            'members': {'some_member': 'maintainer'},
        })
        obj.process()
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_members_not_found(self, cass):
        # Clean up
        GitLabracadabraUser._USERS_USERNAME2ID = {}
        GitLabracadabraUser._USERS_ID2USERNAME = {}
        obj = GitLabracadabraGroup('memory', 'test/group_members', {
            'members': {'member_not_found': 'maintainer'},
            'unknown_members': 'ignore',
        })
        with patch('gitlabracadabra.mixins.members.logger', autospec=True) as logger:
            obj.process()
            self.assertTrue(cass.all_played)
            logger.warning.assert_called_once_with('[%s] User not found %s',
                                                   'test/group_members', 'member_not_found')

    @my_vcr.use_cassette
    def test_members_delete_unknown(self, cass):
        obj = GitLabracadabraGroup('memory', 'test/group_members', {
            'members': {},
            'unknown_members': 'delete',
        })
        obj.process()
        self.assertTrue(cass.all_played)

    # https://github.com/python-gitlab/python-gitlab/pull/1139
    @skipIf(gitlab_version in ['1.6.0', '2.4.0'], 'python-gitlab without share group with groups')
    @my_vcr.use_cassette
    def test_groups(self, cass):
        obj = GitLabracadabraGroup('memory', 'test/group_of_groups1', {
            'groups': {'test/group1': 'developer'},
        })
        obj.process()
        self.assertTrue(cass.all_played)

    # https://github.com/python-gitlab/python-gitlab/pull/1139
    @skipIf(gitlab_version in ['1.6.0', '2.4.0'], 'python-gitlab without share group with groups')
    @my_vcr.use_cassette
    def test_groups_change_access_level(self, cass):
        obj = GitLabracadabraGroup('memory', 'test/group_of_groups2', {
            'groups': {'test/group1': 'maintainer'},
        })
        obj.process()
        self.assertTrue(cass.all_played)

    # https://github.com/python-gitlab/python-gitlab/pull/1139
    @skipIf(gitlab_version in ['1.6.0', '2.4.0'], 'python-gitlab without share group with groups')
    @my_vcr.use_cassette
    def test_groups_not_found(self, cass):
        obj = GitLabracadabraGroup('memory', 'test/group_of_groups3', {
            'groups': {'test/group_not_found': 'maintainer'},
            'unknown_groups': 'ignore',
        })
        with patch('gitlabracadabra.mixins.groups.logger', autospec=True) as logger:
            obj.process()
            self.assertTrue(cass.all_played)
            logger.warning.assert_called_once_with('[%s] Group not found %s',
                                                   'test/group_of_groups3',
                                                   'test/group_not_found')

    # https://github.com/python-gitlab/python-gitlab/pull/1139
    @skipIf(gitlab_version in ['1.6.0', '2.4.0'], 'python-gitlab without share group with groups')
    @my_vcr.use_cassette
    def test_groups_delete_unknown(self, cass):
        obj = GitLabracadabraGroup('memory', 'test/group_of_groups4', {
            'groups': {},
            'unknown_groups': 'delete',
        })
        obj.process()
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_variables(self, cass):
        obj = GitLabracadabraGroup('memory', 'test/group_variables', {
            'variables': [
                {
                    'key': 'simple_var',
                    'value': 'simple_value',
                },
                {
                    'key': 'file_variable',
                    'value': 'BEGIN CERTIFICATE',
                    'variable_type': 'file',
                },
                {
                    'key': 'change_me',
                    'value': 'new_value',
                    'masked': False,
                    'protected': False,
                    'variable_type': 'file',
                },
            ],
            'unknown_variables': 'ignore',
        })
        obj.process()
        self.assertTrue(cass.all_played)

    # https://github.com/python-gitlab/python-gitlab/pull/847
    @skipIf(gitlab_version in ['1.6.0', '1.10.0', '1.11.0', '1.13.0'], 'python-gitlab without group labels support')
    @my_vcr.use_cassette
    def test_labels(self, cass):
        obj = GitLabracadabraGroup('memory', 'test/group_labels', {
            'labels': [
                {
                    'name': 'new_group_label',
                    'color': 'red',
                },
                {
                    'name': 'change_this_group_label',
                    'color': 'green',
                    'description': 'New description',
                },
            ],
            'unknown_labels': 'delete',
        })
        obj.process()
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_milestones(self, cass):
        obj = GitLabracadabraGroup('memory', 'test/group_milestones', {
            'milestones': [
                {
                    'title': 'new_milestone',
                    'description': 'New milestone',
                    'due_date': '2023-01-23',
                    'start_date': '2022-01-23',
                    # 'state': 'active',  # FIXME
                },
                {
                    'title': 'existing_milestone',
                    'description': 'New description',
                    'due_date': '',
                    'start_date': '',
                    'state': 'closed',
                },
            ],
            'unknown_milestones': 'delete',
        })
        obj.process()
        self.assertTrue(cass.all_played)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gitlabracadabra.objects.object import GitLabracadabraObject


class GitLabracadabraUser(GitLabracadabraObject):
    EXAMPLE_YAML_HEADER = 'mmyuser:\n  type: user\n'
    DOC = [
        '# User lifecycle',
        'gitlab_id',
        'create_object',
        'delete_object',

        '# Edit',
        '## Account',
        'name',
        # 'username',
        'email',
        'skip_confirmation',
        'skip_reconfirmation',
        'public_email',
        '## Password',
        'password',
        'reset_password',
        '## Access',
        'projects_limit',
        'can_create_group',
        'admin',
        'external',
        'provider',
        'extern_uid',
        '## Limits',
        'shared_runners_minutes_limit',
        'extra_shared_runners_minutes_limit',
        '## Profile',
        'avatar',
        'skype',
        'linkedin',
        'twitter',
        'website_url',
        'location',
        'organization',
        'bio',
        'private_profile',
        'note',
    ]
    SCHEMA = {
        '$schema': 'http://json-schema.org/draft-04/schema#',
        'title': 'User',
        'type': 'object',
        'properties': {
            # Standard properties
            'gitlab_id': {
                'type': 'string',
                'description': 'GitLab id',
                '_example': 'gitlab',
                '_doc_link': 'action_file.md#gitlab_id',
            },
            'create_object': {
                'type': 'boolean',
                'description': 'Create object if it does not exists',
            },
            'delete_object': {
                'type': 'boolean',
                'description': 'Delete object if it exists',
            },
            # From https://docs.gitlab.com/ee/api/users.html#user-creation
            # 'username': {
            #     'type': 'string',
            #     'description': 'Username',
            # },
            'name': {
                'type': 'string',
                'description': 'Name',
            },
            'email': {
                'type': 'string',
                'description': 'Email',
            },
            'skip_confirmation': {
                'type': 'boolean',
                'description': 'Skip confirmation and assume e-mail is verified',
            },
            'skip_reconfirmation': {
                'type': 'boolean',
                'description': 'Skip reconfirmation',
            },
            'public_email': {
                'type': 'string',
                'description': 'The public email of the user',
            },
            'password': {
                'type': 'string',
                'description': 'Password',
            },
            'reset_password': {
                'type': 'boolean',
                'description': 'Send user password reset link',
            },
            'projects_limit': {
                'type': 'integer',
                'description': 'Number of projects user can create',
                'multipleOf': 1,
                'minimum': 0,
            },
            'can_create_group': {
                'type': 'boolean',
                'description': 'User can create groups',
            },
            'admin': {
                'type': 'boolean',
                'description': 'User is admin',
            },
            'external': {
                'type': 'boolean',
                'description': 'Flags the user as external',
            },
            'provider': {
                'type': 'string',
                'description': 'External provider name',
            },
            'extern_uid': {
                'type': 'string',
                'description': 'External UID',
            },
            'shared_runners_minutes_limit': {
                'type': 'integer',
                'description': 'Pipeline minutes quota for this user',
                'multipleOf': 1,
                'minimum': 0,
            },
            'extra_shared_runners_minutes_limit': {
                'type': 'integer',
                'description': 'Extra pipeline minutes quota for this user',
                'multipleOf': 1,
                'minimum': 0,
            },
            'avatar': {
                'type': 'string',
                'description': 'Image file for user’s avatar',
            },
            'skype': {
                'type': 'string',
                'description': 'Skype ID',
            },
            'linkedin': {
                'type': 'string',
                'description': 'LinkedIn',
            },
            'twitter': {
                'type': 'string',
                'description': 'Twitter account',
            },
            'website_url': {
                'type': 'string',
                'description': 'Website URL',
            },
            'location': {
                'type': 'string',
                'description': 'User’s location',
            },
            'organization': {
                'type': 'string',
                'description': 'Organization name',
            },
            'bio': {
                'type': 'string',
                'description': 'User’s biography',
            },
            'private_profile': {
                'type': 'boolean',
                'description': 'User’s profile is private',
            },
            'note': {
                'type': 'string',
                'description': 'Admin note',
            },
        },
        'additionalProperties': False,
        'dependencies': {
            'email': ['skip_reconfirmation'],
        },
    }

    FIND_PARAM = 'username'

    CREATE_KEY = 'username'

    CREATE_PARAMS = ['email', 'password', 'reset_password', 'skip_confirmation', 'name']

    IGNORED_PARAMS = ['password', 'reset_password', 'skip_confirmation', 'skip_reconfirmation']

    """"_get_param()

    Get a param value.
    """
    def _get_param(self, param_name):
        if param_name == 'admin':
            param_name = 'is_admin'
        return super()._get_param(param_name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from copy import deepcopy


logger = logging.getLogger(__name__)


class MilestonesMixin(object):
    """Object with milestones."""

    """_process_milestones()

    Process the milestones param.
    """
    def _process_milestones(self, param_name, param_value, dry_run=False, skip_save=False):
        assert param_name == 'milestones'  # noqa: S101
        assert not skip_save  # noqa: S101
        unknown_milestones = self._content.get('unknown_milestones', 'warn')
        try:
            current_milestones = dict([[current_milestone.title, current_milestone]
                                      for current_milestone in self._obj.milestones.list(all=True)])
        except AttributeError:
            # https://github.com/python-gitlab/python-gitlab/pull/847
            logger.error('[%s] Unable to manage milestones: %s',
                         self._name, 'group milestones requires python-gitlab >= 1.1.0')
            return
        target_milestones = dict([[target_milestone['title'], deepcopy(target_milestone)]
                                 for target_milestone in param_value])
        # We first check for already existing milestones
        for current_milestone_title, current_milestone in sorted(current_milestones.items()):
            if current_milestone_title in target_milestones:
                for target_milestone_param_name, target_milestone_param_value in (
                    sorted(target_milestones[current_milestone_title].items())
                ):
                    try:
                        current_milestone_param_value = getattr(current_milestone, target_milestone_param_name)
                    except AttributeError:
                        logger.info('[%s] NOT Changing milestone %s %s: %s -> %s (current value is not available)',
                                    self._name, current_milestone_title, target_milestone_param_name,
                                    None, target_milestone_param_value)
                        continue
                    if current_milestone_param_value is None:
                        current_milestone_param_value = ''
                    if current_milestone_param_value != target_milestone_param_value:
                        if target_milestone_param_name == 'state':
                            target_milestone_param_name = 'state_event'
                            if target_milestone_param_value == 'closed':
                                target_milestone_param_value = 'close'
                            else:
                                target_milestone_param_value = 'activate'
                        if dry_run:
                            logger.info('[%s] NOT Changing milestone %s %s: %s -> %s (dry-run)',
                                        self._name, current_milestone_title, target_milestone_param_name,
                                        current_milestone_param_value, target_milestone_param_value)
                        else:
                            logger.info('[%s] Changing milestone %s %s: %s -> %s',
                                        self._name, current_milestone_title, target_milestone_param_name,
                                        current_milestone_param_value, target_milestone_param_value)
                            setattr(current_milestone, target_milestone_param_name, target_milestone_param_value)
                            current_milestone.save()
                target_milestones.pop(current_milestone_title)
            else:
                if unknown_milestones in ['delete', 'remove']:
                    if dry_run:
                        logger.info('[%s] NOT Removing milestone %s (dry-run)',
                                    self._name, current_milestone_title)
                    else:
                        logger.info('[%s] Removing milestone %s',
                                    self._name, current_milestone_title)
                        current_milestone.delete()
                elif unknown_milestones not in ['ignore', 'skip']:
                    logger.warning('[%s] NOT Removing milestone: %s',
                                   self._name, current_milestone_title)
        # Remaining milestones
        for target_milestone_title, target_milestone in sorted(target_milestones.items()):
            if dry_run:
                logger.info('[%s] NOT Adding milestone %s: %s -> %s (dry-run)',
                            self._name, target_milestone_title, None, target_milestone)
            else:
                logger.info('[%s] Adding milestone %s: %s -> %s',
                            self._name, target_milestone_title, None, target_milestone)
                self._obj.milestones.create(target_milestone)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2021 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from __future__ import annotations

from pygit2 import UserPass

from gitlabracadabra.gitlab.pygitlab import PyGitlab


class PyGit2(PyGitlab):
    """PyGit2 wrapper."""

    @property
    def pygit2_credentials(self) -> UserPass:
        """Get PyGit2 credentials.

        Returns:
            A pygit2.UserPass.

        Raises:
            ValueError: No PyGit2 credentials.
        """
        if self._gl.private_token:
            return UserPass('oauth2', self.pygitlab.private_token)
        if self.pygitlab.oauth_token:
            return UserPass('oauth2', self.pygitlab.oauth_token)
        if self.pygitlab.job_token:
            return UserPass('gitlab-ci-token', self.pygitlab.job_token)
        if self.pygitlab.http_username:
            return UserPass(self.pygitlab.http_username, self.pygitlab.http_password)
        raise ValueError('No PyGit2 credentials')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2021 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import namedtuple
from re import search as re_search, sub as re_sub
from typing import Dict, Optional, Tuple, Union

from gitlabracadabra.auth_info import AuthInfo
from gitlabracadabra.containers.const import DOCKER_HOSTNAME, DOCKER_REGISTRY
from gitlabracadabra.containers.manifest import Manifest
from gitlabracadabra.containers.registry import Registry
from gitlabracadabra.singleton import SingletonMeta


ReferenceParts = namedtuple('ReferenceParts', ['hostname', 'manifest_name', 'tag', 'digest'])


class Registries(object, metaclass=SingletonMeta):  # noqa: WPS214
    """All registies by name."""

    def __init__(self) -> None:
        """All connected registries.

        Intented to be used as a singleton.
        """
        self._registries: Dict[str, Registry] = {}

    def reset(self) -> None:
        """Reset registry cache."""
        self._registries = {}

    def get_registry(self, hostname: str, auth_info: Optional[AuthInfo] = None) -> Registry:
        """Get a registry connection.

        Args:
            hostname: fqdn of a registry.
            auth_info: Authentication information.

        Returns:
            The registry with the given hostname
        """
        if hostname == DOCKER_REGISTRY:
            hostname = DOCKER_HOSTNAME
        if hostname not in self._registries:
            self._registries[hostname] = Registry(hostname, auth_info)
        return self._registries[hostname]

    def get_manifest(self, name: Union[str, ReferenceParts]) -> Manifest:
        """Get a manifest.

        Args:
            name: Reference name, or reference parts.

        Returns:
            The Manifest with the given full reference name.
        """
        if isinstance(name, str):
            full_reference_parts = self.full_reference_parts(name)
        else:
            full_reference_parts = name
        registry = self.get_registry(full_reference_parts.hostname)
        return Manifest(
            registry,
            full_reference_parts.manifest_name,
            full_reference_parts.digest,
            tag=full_reference_parts.tag,
        )

    @classmethod
    def short_reference(cls, name: str) -> str:
        """Get short reference (i.e. familiar name).

        Args:
            name: Reference name.

        Returns:
            The corresponding short reference name.
        """
        short_reference = cls.full_reference(name)
        if short_reference.startswith('{0}/library/'.format(DOCKER_HOSTNAME)):
            prefix_len = len(DOCKER_HOSTNAME) + 1 + len('library') + 1
            short_reference = short_reference[prefix_len:]
        if short_reference.startswith('{0}/'.format(DOCKER_HOSTNAME)):
            prefix_len = len(DOCKER_HOSTNAME) + 1
            short_reference = short_reference[prefix_len:]
        return re_sub(':latest(@sha256:[0-9A-Fa-f]{64})?$', r'\1', short_reference)

    @classmethod
    def full_reference(cls, name: str) -> str:
        """Get full reference.

        Args:
            name: Reference name.

        Returns:
            The corresponding full reference name.
        """
        full_reference_parts = cls.full_reference_parts(name)
        full_reference = '{0}/{1}'.format(full_reference_parts.hostname, full_reference_parts.manifest_name)
        if full_reference_parts.tag:
            full_reference = '{0}:{1}'.format(full_reference, full_reference_parts.tag)
        if full_reference_parts.digest:
            full_reference = '{0}@{1}'.format(full_reference, full_reference_parts.digest)
        return full_reference

    @classmethod
    def full_reference_parts(cls, name: str) -> ReferenceParts:
        """Get full reference parts (hostname, manifest_name, tag, digest).

        Args:
            name: Reference name.

        Returns:
            The corresponding full reference parts.
        """
        hostname, remaining = cls._split_docker_domain(name)
        digest: Optional[str]
        tag: Optional[str]
        try:
            remaining, digest = remaining.split('@', 1)
        except ValueError:
            digest = None
        try:
            remaining, tag = remaining.split(':', 1)
        except ValueError:
            tag = None
        if hostname == DOCKER_HOSTNAME and '/' not in remaining:
            remaining = 'library/{0}'.format(remaining)
        return ReferenceParts(hostname, remaining, tag, digest)

    @classmethod
    def _split_docker_domain(cls, name: str) -> Tuple[str, str]:
        parts = name.split('/', 1)
        if len(parts) == 2 and re_search(r'^localhost$|:\d|\.', parts[0]):
            return parts[0], parts[1]
        return DOCKER_HOSTNAME, name



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2021 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from gitlab.exceptions import GitlabCreateError, GitlabListError

from gitlabracadabra.gitlab.access_levels import access_level
from gitlabracadabra.mixins.boards import BoardsMixin
from gitlabracadabra.mixins.groups import GroupsMixin
from gitlabracadabra.mixins.image_mirrors import ImageMirrorsMixin
from gitlabracadabra.mixins.labels import LabelsMixin
from gitlabracadabra.mixins.members import MembersMixin
from gitlabracadabra.mixins.milestones import MilestonesMixin
from gitlabracadabra.mixins.mirrors import MirrorsMixin
from gitlabracadabra.mixins.pipeline_schedules import PipelineSchedulesMixin
from gitlabracadabra.mixins.rename_branches import RenameBranchesMixin
from gitlabracadabra.mixins.variables import VariablesMixin
from gitlabracadabra.mixins.webhooks import WebhooksMixin
from gitlabracadabra.objects.object import GitLabracadabraObject


logger = logging.getLogger(__name__)


class GitLabracadabraProject(GitLabracadabraObject, BoardsMixin, GroupsMixin, ImageMirrorsMixin, LabelsMixin,
                             MembersMixin, MilestonesMixin, MirrorsMixin, PipelineSchedulesMixin, RenameBranchesMixin,
                             VariablesMixin, WebhooksMixin):
    EXAMPLE_YAML_HEADER = 'mygroup/myproject:\n'
    DOC = [
        '# Project lifecycle',
        'gitlab_id',
        'create_object',
        'delete_object',
        'archived',
        'initialize_with_readme',

        '# General Settings',
        '## Naming, topics, avatar',
        'name',
        'tag_list',
        'description',
        # 'avatar',  # FIXME Gitlabracadabra
        '## Visibility, project features, permissions',
        'visibility',
        'request_access_enabled',
        'issues_access_level',
        'repository_access_level',
        'merge_requests_access_level',
        'forking_access_level',
        'builds_access_level',
        'container_registry_enabled',
        'lfs_enabled',
        # 'metrics_dashboard_access_level',  # FIXME Gitlab
        'packages_enabled',
        'wiki_access_level',
        'snippets_access_level',
        'pages_access_level',
        'emails_disabled',
        # 'show_default_award_emojis',  # FIXME Gitlab
        '## Merge requests',
        'merge_method',
        # 'merge_pipelines_enabled',  # FIXME Gitlab
        'resolve_outdated_diff_discussions',
        'printing_merge_request_link_enabled',
        'remove_source_branch_after_merge',
        'only_allow_merge_if_pipeline_succeeds',
        'allow_merge_on_skipped_pipeline',
        'only_allow_merge_if_all_discussions_are_resolved',
        'suggestion_commit_message',
        # 'merge_requests_template',  # FIXME
        '## Merge request approvals',  # FIXME ...
        'approvals_before_merge',
        # '## Badges',  # FIXME ...
        # '## Default issue template',  # FIXME ...
        # '## Service Desk',  # FIXME ...

        '# Members',
        'members',
        'unknown_members',
        'groups',
        'unknown_groups',

        # '# Integrations',  # FIXME

        '# Webhooks',
        'webhooks',
        'unknown_webhooks',

        '# Repository Settings',
        '## Default branch',
        'default_branch',
        'autoclose_referenced_issues',
        # '## Push Rules',  # FIXME ...
        '## Mirroring repositories',
        'mirrors',
        'mirror',
        'import_url',
        'mirror_user_id',
        'mirror_overwrites_diverged_branches',
        'mirror_trigger_builds',
        'only_mirror_protected_branches',
        '## Protected Branches',
        'protected_branches',
        'unknown_protected_branches',
        '## Protected Tags',
        'protected_tags',
        'unknown_protected_tags',
        # '## Deploy Keys',  # FIXME ...
        # '## Deploy Tokens',  # FIXME ...
        '## Branches',
        'branches',
        'rename_branches',

        '# CI / CD Settings',
        '## General pipelines',
        'build_git_strategy',
        'ci_default_git_depth',
        'build_timeout',
        'ci_config_path',
        'public_builds',
        'auto_cancel_pending_pipelines',
        # 'forward_deployment_enabled',  # FIXME Gitlab
        'build_coverage_regex',
        '## Auto DevOps',
        'auto_devops_enabled',
        'auto_devops_deploy_strategy',
        # '## Protected Environments',  # FIXME ...
        '## Runners',
        'shared_runners_enabled',
        '## Variables',
        'variables',
        'unknown_variables',
        '## Clean up image tags',
        'container_expiration_policy',

        '# Issues',
        '## Boards',
        'boards',
        'unknown_boards',
        'unknown_board_lists',
        '## Labels',
        'labels',
        'unknown_labels',
        '## Milestones',
        'milestones',
        'unknown_milestones',

        '# CI/CD',
        '## Pipeline schedules',
        'pipeline_schedules',
        'unknown_pipeline_schedules',
        'unknown_pipeline_schedule_variables',

        '# Mirroring container images',
        'image_mirrors',

        '# Deprecated',
        'issues_enabled',
        'merge_requests_enabled',
        'jobs_enabled',
        'wiki_enabled',
        'snippets_enabled',
    ]
    SCHEMA = {
        '$schema': 'http://json-schema.org/draft-04/schema#',
        'title': 'Project',
        'type': 'object',
        'properties': {
            # Standard properties
            'gitlab_id': {
                'type': 'string',
                'description': 'GitLab id',
                '_example': 'gitlab',
                '_doc_link': 'action_file.md#gitlab_id',
            },
            'create_object': {
                'type': 'boolean',
                'description': 'Create object if it does not exists',
            },
            'delete_object': {
                'type': 'boolean',
                'description': 'Delete object if it exists',
            },
            # https://docs.gitlab.com/ee/api/projects.html#archive-a-project
            # https://docs.gitlab.com/ee/api/projects.html#unarchive-a-project
            'archived': {
                'type': 'boolean',
                'description': 'Archive or unarchive project',
            },
            # From https://docs.gitlab.com/ee/api/projects.html#create-project
            # and https://docs.gitlab.com/ee/api/projects.html#edit-project
            'initialize_with_readme': {
                'type': 'boolean',
                'description': 'false by default',
            },
            'name': {
                'type': 'string',
                'description': 'Project name',
            },
            # 'path': {
            #     'type': 'string',
            #     'description': 'Repository name for new project. '
            #                    'Generated based on name if not provided (generated lowercased with dashes).',
            # },
            'tag_list': {
                'type': 'array',
                'description': 'Topics',
                'items': {
                    'type': 'string',
                },
                'uniqueItems': True,
                '_example': '[GitLab, API, YAML]',
            },
            'description': {
                'type': 'string',
                'description': 'Project description',
                '_example': '|-\n    🧹 GitLabracadabra 🧙\n\n    :alembic: Adds some magic to GitLab :crystal\\_ball:',
            },
            # 'avatar': {
            #     'type': 'string',
            #     'description': 'Project avatar',
            # },
            'visibility': {
                'type': 'string',
                'description': 'Project visibility',
                'enum': ['private', 'internal', 'public'],
            },
            'request_access_enabled': {
                'type': 'boolean',
                'description': 'Allow users to request access',
            },
            'issues_access_level': {
                'type': 'string',
                'description': 'Issues access level.',
                'enum': ['disabled', 'private', 'enabled'],
            },
            'repository_access_level': {
                'type': 'string',
                'description': 'Repository access level.',
                'enum': ['disabled', 'private', 'enabled'],
            },
            'merge_requests_access_level': {
                'type': 'string',
                'description': 'Merge requests access level.',
                'enum': ['disabled', 'private', 'enabled'],
            },
            'forking_access_level': {
                'type': 'string',
                'description': 'Forking access level.',
                'enum': ['disabled', 'private', 'enabled'],
            },
            'builds_access_level': {
                'type': 'string',
                'description': 'Builds access level.',
                'enum': ['disabled', 'private', 'enabled'],
            },
            'container_registry_enabled': {
                'type': 'boolean',
                'description': 'Enable container registry for this project',
            },
            'lfs_enabled': {
                'type': 'boolean',
                'description': 'Enable LFS',
            },
            'packages_enabled': {
                'type': 'boolean',
                'description': 'Enable or disable packages repository feature',
            },
            'wiki_access_level': {
                'type': 'string',
                'description': 'Wiki access level.',
                'enum': ['disabled', 'private', 'enabled'],
            },
            'snippets_access_level': {
                'type': 'string',
                'description': 'Snippets access level.',
                'enum': ['disabled', 'private', 'enabled'],
            },
            'pages_access_level': {
                'type': 'string',
                'description': 'Forking access level.',
                'enum': ['disabled', 'private', 'enabled', 'public'],
            },
            'emails_disabled': {
                'type': 'boolean',
                'description': 'Disable email notifications',
            },
            'merge_method': {
                'type': 'string',
                'description': 'Set the merge method used',
                'enum': ['merge', 'rebase_merge', 'ff'],
            },
            'resolve_outdated_diff_discussions': {
                'type': 'boolean',
                'description': 'Automatically resolve merge request diffs discussions on lines changed with a push',
            },
            'printing_merge_request_link_enabled': {
                'type': 'boolean',
                'description': 'Show link to create/view merge request when pushing from the command line',
            },
            'remove_source_branch_after_merge': {
                'type': 'boolean',
                'description': 'Enable Delete source branch option by default for all new merge requests',
            },
            'only_allow_merge_if_pipeline_succeeds': {
                'type': 'boolean',
                'description': 'Set whether merge requests can only be merged with successful jobs',
            },
            'allow_merge_on_skipped_pipeline': {
                'type': 'boolean',
                'description': 'Set whether or not merge requests can be merged with skipped jobs',
            },
            'only_allow_merge_if_all_discussions_are_resolved': {
                'type': 'boolean',
                'description': 'Set whether merge requests can only be merged when all the discussions are resolved',
            },
            'suggestion_commit_message': {
                'type': 'string',
                'description': 'The commit message used to apply merge request suggestions',
            },
            'approvals_before_merge': {
                'type': 'integer',
                'description': 'How many approvers should approve merge request by default',
                'multipleOf': 1,
                'minimum': 0,
            },
            # From https://docs.gitlab.com/ee/api/members.html#add-a-member-to-a-group-or-project
            # FIXME expires_at not supported
            'members': {
                'type': 'object',
                'description': 'Members',
                'additionalProperties': {
                    'type': 'string',
                    'description': 'The permissions level to grant the member.',
                    'enum': ['guest', 'reporter', 'developer', 'maintainer', 'owner'],
                },
                '_example': ('\n'
                             '    foo: developer\n'
                             '    bar: maintainer # one of guest, reporter, developer, maintainer, owner\n'),
            },
            'unknown_members': {  # GitLabracadabra
                'type': 'string',
                'description': 'What to do with unknown members (`warn` by default).',
                'enum': ['warn', 'delete', 'remove', 'ignore', 'skip'],
            },
            # From https://docs.gitlab.com/ee/api/projects.html#share-project-with-group_access_level
            # and https://docs.gitlab.com/ee/api/projects.html#delete-a-shared-project-link-within-a-group
            # FIXME expires_at not supported
            'groups': {
                'type': 'object',
                'description': 'Groups',
                'additionalProperties': {
                    'type': 'string',
                    'description': 'The permissions level to grant the group.',
                    'enum': ['guest', 'reporter', 'developer', 'maintainer'],
                },
                '_example': ('\n'
                             '    group/foo: guest\n'
                             '    group/bar: reporter # one of guest, reporter, developer, maintainer\n'),
            },
            'unknown_groups': {  # GitLabracadabra
                'type': 'string',
                'description': 'What to do with unknown groups (`warn` by default).',
                'enum': ['warn', 'delete', 'remove', 'ignore', 'skip'],
            },
            # From https://docs.gitlab.com/ee/api/projects.html#hooks
            'webhooks': {
                'type': 'array',
                'description': "The list of project's webhooks",
                'items': {
                    'type': 'object',
                    'properties': {
                        'url': {
                            'type': 'string',
                            'description': 'The hook URL',
                        },
                        'push_events': {
                            'type': 'boolean',
                            'description': 'Trigger hook on push events',
                        },
                        'push_events_branch_filter': {
                            'type': 'string',
                            'description': 'Trigger hook on push events for matching branches only',
                        },
                        'issues_events': {
                            'type': 'boolean',
                            'description': 'Trigger hook on issues events',
                        },
                        'confidential_issues_events': {
                            'type': 'boolean',
                            'description': 'Trigger hook on confidential issues events',
                        },
                        'merge_requests_events': {
                            'type': 'boolean',
                            'description': 'Trigger hook on merge requests events',
                        },
                        'tag_push_events': {
                            'type': 'boolean',
                            'description': 'Trigger hook on tag push events',
                        },
                        'note_events': {
                            'type': 'boolean',
                            'description': 'Trigger hook on note events',
                        },
                        'confidential_note_events': {
                            'type': 'boolean',
                            'description': 'Trigger hook on confidential note events',
                        },
                        'job_events': {
                            'type': 'boolean',
                            'description': 'Trigger hook on job events',
                        },
                        'pipeline_events': {
                            'type': 'boolean',
                            'description': 'Trigger hook on pipeline events',
                        },
                        'wiki_page_events': {
                            'type': 'boolean',
                            'description': 'Trigger hook on wiki events',
                        },
                        'enable_ssl_verification': {
                            'type': 'boolean',
                            'description': 'Do SSL verification when triggering the hook',
                        },
                        'token': {
                            'type': 'string',
                            'description': ('Secret token to validate received payloads; '
                                            'this will not be returned in the response'),
                        },
                    },
                    'required': ['url'],
                    'additionalProperties': False,
                },
                'uniqueItems': True,
                '_example': ('\n'
                             '    - url: http://example.com/api/trigger\n'
                             '      push_events: true\n'
                             "      push_events_branch_filter: ''\n"
                             '      issues_events: true\n'
                             '      confidential_issues_events: true\n'
                             '      merge_requests_events: true\n'
                             '      tag_push_events: true\n'
                             '      note_events: true\n'
                             '      confidential_note_events: true\n'
                             '      job_events: true\n'
                             '      pipeline_events: true\n'
                             '      wiki_page_events: true\n'
                             '      enable_ssl_verification: true\n'
                             '      # token: T0k3N\n'),
            },
            'unknown_webhooks': {  # GitLabracadabra
                'type': 'string',
                'description': 'What to do with unknown webhooks (`warn` by default).',
                'enum': ['warn', 'delete', 'remove', 'ignore', 'skip'],
            },
            # From https://docs.gitlab.com/ee/api/projects.html#edit-project
            'default_branch': {
                'type': 'string',
                'description': '`master` by default',
            },
            'autoclose_referenced_issues': {
                'type': 'boolean',
                'description': 'Set whether auto-closing referenced issues on default branch',
            },
            # GitLabracadabra
            'mirrors': {
                'type': 'array',
                'description': "The list of project's mirrors",
                'items': {
                    'type': 'object',
                    'properties': {
                        'url': {
                            'type': 'string',
                            'description': 'Repository URL',
                        },
                        'auth_id': {
                            'type': 'string',
                            'description': 'Section from .python-gitlab.cfg for authentication',
                        },
                        'direction': {
                            'type': 'string',
                            'description': 'Mirror direction',
                            'enum': ['pull', 'push'],
                        },
                        'skip_ci': {
                            'type': 'boolean',
                            'description': 'Skip CI during push',
                        },
                        'branches': {
                            'type': 'array',
                            'description': 'The branches mapping',
                            'items': {
                                'type': 'object',
                                'properties': {
                                    'from': {
                                        'type': 'string',
                                        'description': 'Source name or regular expression',
                                    },
                                    'to': {
                                        'type': 'string',
                                        'description': 'Destination name or regular expression template',
                                    },
                                },
                                'required': ['from'],
                                'additionalProperties': False,
                            },
                            'uniqueItems': True,
                        },
                        'tags': {
                            'type': 'array',
                            'description': 'The tags mapping',
                            'items': {
                                'type': 'object',
                                'properties': {
                                    'from': {
                                        'type': 'string',
                                        'description': 'Source name or regular expression',
                                    },
                                    'to': {
                                        'type': 'string',
                                        'description': 'Destination name or regular expression template',
                                    },
                                },
                                'required': ['from'],
                                'additionalProperties': False,
                            },
                            'uniqueItems': True,
                        },
                    },
                    'required': ['url'],
                    'additionalProperties': False,
                },
                'uniqueItems': True,
                '_example': ('\n'
                             '    - url: https://gitlab.com/gitlabracadabra/gitlabracadabra.git\n'
                             '      # auth_id: gitlab # Section from .python-gitlab.cfg for authentication\n'
                             '      direction: pull # one of pull, push ; only first pull mirror is processed\n'
                             # '      skip_ci: true # defaults to true\n'
                             '      branches: # if you omit this parameter, all branches are mirrored\n'
                             "        - from: '/wip-.*/'\n"
                             "          to: '' # This will skip those branches\n"
                             '        - from: master\n'
                             '          # to: master # implicitly equal to source branch\n'
                             # '          # skip_ci: false # inherited by default \n'
                             # '          merge_request: true  '
                             # '# Will create a temporary branch and create '
                             # 'a MR targeting master (the `to` parameter)\n'
                             # '          merge_request_from_fork: myself/myproject # will fork if needed\n'
                             '        # Using regexps\n'
                             "        - from: '/(.*)/'\n"
                             "          to: 'upstream/\\1'\n"
                             # '          skip_ci: true\n'
                             # '          # force_push: false\n'
                             '      tags: # if you omit this parameter, all tags are mirrored\n'
                             "        - from: '/v(.*)/i'\n"
                             "          to: 'upstream-\\1'\n"
                             # '          skip_ci: true\n'
                             '  builds_access_level: disabled # If you want to prevent triggering pipelines on push'),
                '_doc_link': 'mirrors.md',
            },
            # From https://docs.gitlab.com/ee/api/projects.html#edit-project
            'mirror': {
                'type': 'boolean',
                'description': 'Enables pull mirroring in a project',
            },
            'import_url': {
                'type': 'string',
                'description': 'URL to import repository from',
            },
            'mirror_user_id': {
                'type': 'integer',
                'description': 'User responsible for all the activity surrounding a pull mirror event',
            },
            'mirror_overwrites_diverged_branches': {
                'type': 'boolean',
                'description': 'Pull mirror overwrites diverged branches',
            },
            'mirror_trigger_builds': {
                'type': 'boolean',
                'description': 'Pull mirroring triggers builds',
            },
            'only_mirror_protected_branches': {
                'type': 'boolean',
                'description': 'Only mirror protected branches',
            },
            # From https://docs.gitlab.com/ee/api/protected_branches.html#protect-repository-branches
            # FIXME EE features: unprotect_access_level, allowed_to_push, allowed_to_merge, allowed_to_unprotect
            'protected_branches': {
                'type': 'object',
                'description': 'Protected branches',
                'additionalProperties': {
                    'type': 'object',
                    'properties': {
                        'push_access_level': {
                            'type': 'string',
                            'description': 'Access levels allowed to push (defaults: maintainer access level)',
                            'enum': ['noone', 'developer', 'maintainer'],
                        },
                        'merge_access_level': {
                            'type': 'string',
                            'description': 'Access levels allowed to merge (defaults: maintainer access level)',
                            'enum': ['noone', 'developer', 'maintainer'],
                        },
                    },
                },
                '_example': ('\n'
                             '    master:\n'
                             '      merge_access_level: maintainer # one of noone, developer, maintainer\n'
                             '      push_access_level: noone\n'
                             '    develop:\n'
                             '      merge_access_level: developer\n'
                             '      push_access_level: noone\n'),
            },
            'unknown_protected_branches': {  # GitLabracadabra
                'type': 'string',
                'description': 'What to do with unknown protected branches (`warn` by default).',
                'enum': ['warn', 'delete', 'remove', 'ignore', 'skip'],
            },
            # From https://docs.gitlab.com/ee/api/protected_tags.html#protect-repository-tags
            'protected_tags': {
                'type': 'object',
                'description': 'Protected tags',
                'additionalProperties': {
                    'type': 'string',
                    'description': 'Access levels allowed to create (defaults: maintainer access level)',
                    'enum': ['noone', 'developer', 'maintainer'],
                },
                '_example': ('\n'
                             '    v*: maintainer # one of noone, developer, maintainer\n'
                             '    *: developer # one of noone, developer, maintainer\n'),
            },
            'unknown_protected_tags': {  # GitLabracadabra
                'type': 'string',
                'description': 'What to do with unknown protected tags (`warn` by default).',
                'enum': ['warn', 'delete', 'remove', 'ignore', 'skip'],
            },
            # From https://docs.gitlab.com/ee/api/branches.html#create-repository-branch
            'branches': {
                'type': 'array',
                'description': 'The list of branches for a project. '
                               'Branches are created in order',
                'items': {
                    'type': 'string',
                },
                'uniqueItems': True,
                '_example': ('\n'
                             '    - master\n'
                             '    - develop'),
            },
            'rename_branches': {
                'type': 'array',
                'description': 'Rename branches of a project. '
                               'Rename pairs (old_name: new_name) are processed in order',
                'items': {
                    'type': 'object',
                    'additionalProperties': {
                        'type': 'string',
                        'description': 'The new branch name.',
                    },
                    'minProperties': 1,
                    'maxProperties': 1,
                },
                'uniqueItems': True,
                '_example': ('\n'
                             '    - old_name: new_name\n'
                             '    # To Rename consecutive branches:\n'
                             '    - branch2: branch3\n'
                             '    - branch1: branch2'),
            },
            # From https://docs.gitlab.com/ee/api/projects.html#edit-project
            'build_git_strategy': {
                'type': 'string',
                'description': 'The Git strategy',
                'enum': ['fetch', 'clone'],
            },
            'ci_default_git_depth': {
                'type': 'integer',
                'description': 'Default number of revisions for shallow cloning',
            },
            'build_timeout': {
                'type': 'integer',
                'description': 'The maximum amount of time in minutes that a job is able run (in seconds)',
            },
            'ci_config_path': {
                'type': 'string',
                'description': 'The path to CI config file',
                '_example': 'debian/salsa-ci.yml',
            },
            'public_builds': {
                'type': 'boolean',
                'description': 'If true, jobs can be viewed by non-project-members',
            },
            'auto_cancel_pending_pipelines': {
                'type': 'string',
                'description': 'Auto-cancel pending pipelines',
                'enum': ['enabled', 'disabled'],
            },
            'build_coverage_regex': {
                'type': 'string',
                'description': 'Test coverage parsing',
            },
            'auto_devops_enabled': {
                'type': 'boolean',
                'description': 'Enable Auto DevOps for this project',
            },
            'auto_devops_deploy_strategy': {
                'type': 'string',
                'description': 'Auto Deploy strategy',
                'enum': ['continuous', 'manual', 'timed_incremental'],
            },
            'shared_runners_enabled': {
                'type': 'boolean',
                'description': 'Enable shared runners for this project',
            },
            # From https://docs.gitlab.com/ee/api/project_level_variables.html#create-variable
            'variables': {
                'type': 'array',
                'description': "The list of project's variables",
                'items': {
                    'type': 'object',
                    'properties': {
                        'key': {
                            'type': 'string',
                            'description': 'The key of a variable.',
                            'pattern': '[a-zA-Z0-9_]+',
                        },
                        'value': {
                            'type': 'string',
                            'description': 'The value of a variable.',
                        },
                        'variable_type': {
                            'type': 'string',
                            'description': 'The type of a variable. Available types are: env_var (default) and file.',
                            'enum': ['env_var', 'file'],
                        },
                        'protected': {
                            'type': 'boolean',
                            'description': 'Whether the variable is protected.',
                        },
                        'masked': {
                            'type': 'boolean',
                            'description': 'Whether the variable is masked.',
                        },
                        'environment_scope': {  # Premium+/Silver+
                            'type': 'string',
                            'description': 'The environment_scope of the variable.',
                        },
                    },
                    'required': ['key', 'value'],
                    'additionalProperties': False,
                },
                'uniqueItems': True,
                '_example': ('\n'
                             '    - key: DAST_DISABLED\n'
                             "      value: '1'\n"
                             '      masked: false\n'
                             '      protected: false\n'
                             '      variable_type: env_var\n'),
            },
            'unknown_variables': {  # GitLabracadabra
                'type': 'string',
                'description': 'What to do with unknown variables (`warn` by default).',
                'enum': ['warn', 'delete', 'remove', 'ignore', 'skip'],
            },
            # From https://docs.gitlab.com/ee/api/projects.html#edit-project
            # container_expiration_policy_attributes
            'container_expiration_policy': {
                'type': 'object',
                'description': 'Update the image cleanup policy for this project',
                'properties': {
                    'enabled': {
                        'type': 'boolean',
                    },
                    'cadence': {
                        'type': 'string',
                    },
                    'keep_n': {
                        'type': 'integer',
                    },
                    'name_regex_keep': {
                        'type': 'string',
                    },
                    'older_than': {
                        'type': 'string',
                    },
                    'name_regex_delete': {
                        'type': 'string',
                    },
                },
                'required': ['enabled'],
                'additionalProperties': False,
                '_example': ('\n'
                             '    enabled: true\n'
                             '    cadence: 7d # 1d, 7d, 14d, 1month, 3month\n'
                             '    keep_n: 10 # 1, 5, 10, 25, 50, 100\n'
                             "    name_regex_keep: '.*master|.*release|release-.*|master-.*'\n"
                             '    older_than: 90d\n'
                             "    name_regex_delete: '.*'\n"),
            },
            # From https://docs.gitlab.com/ee/api/boards.html
            'boards': {
                'type': 'array',
                'description': "The list of project's boards",
                'items': {
                    'type': 'object',
                    'properties': {
                        'name': {
                            'type': 'string',
                            'description': 'The name of the board.',
                        },
                        'old_name': {
                            'type': 'string',
                            'description': 'The previous name of the board.',
                        },
                        'hide_backlog_list': {
                            'type': 'boolean',
                            'description': 'Hide the Open list',
                        },
                        'hide_closed_list': {
                            'type': 'boolean',
                            'description': 'Hide the Closed list',
                        },
                        'lists': {
                            'type': 'array',
                            'description': 'Ordered list of labels',
                            'items': {
                                'type': 'object',
                                'properties': {
                                    'label': {
                                        'type': 'string',
                                        'description': 'The name of a label',
                                    },
                                },
                            },
                        },
                        'unknown_lists': {  # GitLabracadabra
                            'type': 'string',
                            'description': ('What to do with unknown board lists '
                                            '(Value of `unknown_board_lists` by default).'),
                            'enum': ['warn', 'delete', 'remove', 'ignore', 'skip'],
                        },
                    },
                    'required': ['name'],
                    'additionalProperties': False,
                },
                'uniqueItems': True,
                '_example': ('\n'
                             '    - name: My Board\n'
                             '      # old_name: Development # Use this to rename a board\n'
                             '      hide_backlog_list: false\n'
                             '      hide_closed_list: false\n'
                             '      lists:\n'
                             '        - label: TODO\n'
                             '        - label: WIP\n'),
            },
            'unknown_boards': {  # GitLabracadabra
                'type': 'string',
                'description': 'What to do with unknown boards (`warn` by default).',
                'enum': ['warn', 'delete', 'remove', 'ignore', 'skip'],
            },
            'unknown_board_lists': {  # GitLabracadabra
                'type': 'string',
                'description': 'What to do with unknown board lists (`delete` by default).',
                'enum': ['warn', 'delete', 'remove', 'ignore', 'skip'],
            },
            # From https://docs.gitlab.com/ee/api/labels.html#create-a-new-label
            'labels': {
                'type': 'array',
                'description': "The list of project's labels",
                'items': {
                    'type': 'object',
                    'properties': {
                        'name': {
                            'type': 'string',
                            'description': 'The name of the label.',
                        },
                        'color': {
                            'type': 'string',
                            'description': 'The color of the label.',
                        },
                        'description': {
                            'type': 'string',
                            'description': 'The description of the label.',
                        },
                        'priority': {
                            'type': 'integer',
                            'description': 'The priority of the label.',
                        },
                    },
                    'required': ['name'],  # color not required to allow priority override
                    'additionalProperties': False,
                },
                'uniqueItems': True,
                '_example': ('\n'
                             '    - name: critical\n'
                             '      priority: 0\n'
                             '    - name: bug\n'
                             '      priority: 1\n'
                             '    - name: confirmed\n'
                             '      priority: 2\n'),
            },
            'unknown_labels': {  # GitLabracadabra
                'type': 'string',
                'description': 'What to do with unknown labels (`warn` by default).',
                'enum': ['warn', 'delete', 'remove', 'ignore', 'skip'],
            },
            # From https://docs.gitlab.com/ee/api/milestones.html#edit-milestone
            'milestones': {
                'type': 'array',
                'description': "The list of project's milestones",
                'items': {
                    'type': 'object',
                    'properties': {
                        'title': {
                            'type': 'string',
                            'description': 'The title of a milestone',
                        },
                        'description': {
                            'type': 'string',
                            'description': 'The description of a milestone',
                        },
                        'due_date': {
                            'type': 'string',
                            'description': 'The due date of the milestone',
                            'pattern': '^(\\d{4}-\\d{2}-\\d{2})?$',
                        },
                        'start_date': {
                            'type': 'string',
                            'description': 'The start date of the milestone',
                            'pattern': '^(\\d{4}-\\d{2}-\\d{2})?$',
                        },
                        'state': {
                            'type': 'string',
                            'description': 'The state event of the milestone',
                            'enum': ['closed', 'active'],
                        },
                    },
                    'required': ['title'],
                    'additionalProperties': False,
                },
                'uniqueItems': True,
                '_example': ('\n'
                             "    - title: '1.0'\n"
                             '      description: Version 1.0\n'
                             "      due_date: '2021-01-23' # Quotes are mandatory\n"
                             "      start_date: '2020-01-23' # Quotes are mandatory\n"
                             '      state: active # or closed\n'),
            },
            'unknown_milestones': {  # GitLabracadabra
                'type': 'string',
                'description': 'What to do with unknown milestones (`warn` by default).',
                'enum': ['warn', 'delete', 'remove', 'ignore', 'skip'],
            },
            # From https://docs.gitlab.com/ee/api/pipeline_schedules.html#create-a-new-pipeline-schedule
            'pipeline_schedules': {
                'type': 'array',
                'description': "The list of project's pipeline schedules",
                'items': {
                    'type': 'object',
                    'properties': {
                        'description': {
                            'type': 'string',
                            'description': 'The description of pipeline schedule',
                            'pattern': '[a-zA-Z0-9_]+',
                        },
                        'ref': {
                            'type': 'string',
                            'description': 'The branch/tag name will be triggered',
                        },
                        'cron': {
                            'type': 'string',
                            'description': ('The cron (e.g. `0 1 * * *`) '
                                            '([Cron syntax](https://en.wikipedia.org/wiki/Cron))'),
                        },
                        'cron_timezone ': {
                            'type': 'string',
                            'description': ('The timezone supported by `ActiveSupport::TimeZone` '
                                            "(e.g. `Pacific Time (US & Canada)`) (default: `'UTC'`)"),
                        },
                        'active': {
                            'type': 'boolean',
                            'description': 'The activation of pipeline schedule',
                        },
                        # From https://docs.gitlab.com/ee/api/pipeline_schedules.html
                        # #create-a-new-pipeline-schedule-variable
                        'variables': {
                            'type': 'array',
                            'description': "The list of project's variables",
                            'items': {
                                'type': 'object',
                                'properties': {
                                    'key': {
                                        'type': 'string',
                                        'description': 'The key of a variable',
                                        'pattern': '[a-zA-Z0-9_]+',
                                    },
                                    'value': {
                                        'type': 'string',
                                        'description': 'The value of a variable',
                                    },
                                    'variable_type': {
                                        'type': 'string',
                                        'description': ('The type of a variable. '
                                                        'Available types are: env_var (default) and file'),
                                        'enum': ['env_var', 'file'],
                                    },
                                },
                                'required': ['key', 'value'],
                                'additionalProperties': False,
                            },
                            'uniqueItems': True,
                        },
                        'unknown_variables': {  # GitLabracadabra
                            'type': 'string',
                            'description': ('What to do with unknown pipeline schedule variables '
                                            '(Value of `unknown_pipeline_schedule_variables` by default).'),
                            'enum': ['warn', 'delete', 'remove', 'ignore', 'skip'],
                        },
                    },
                    'required': ['description', 'ref', 'cron'],
                    'additionalProperties': False,
                },
                'uniqueItems': True,
                '_example': ('\n'
                             '    - description: Build packages\n'
                             '      ref: master\n'
                             "      cron: '0 1 * * 5'\n"
                             '      # cron_timezone: UTC\n'
                             '      # active: true\n'
                             '      variables:\n'
                             '        - key: MY_VAR\n'
                             '          value: my value\n'
                             '          # variable_type: env_var # or file\n'
                             '      # unknown_variables: warn # one of warn, delete, remove, ignore, skip\n'),
            },
            'unknown_pipeline_schedules': {  # GitLabracadabra
                'type': 'string',
                'description': 'What to do with unknown pipeline schedules (`warn` by default).',
                'enum': ['warn', 'delete', 'remove', 'ignore', 'skip'],
            },
            'unknown_pipeline_schedule_variables': {  # GitLabracadabra
                'type': 'string',
                'description': 'What to do with unknown pipeline schedule variables (`warn` by default).',
                'enum': ['warn', 'delete', 'remove', 'ignore', 'skip'],
            },
            # GitLabracadabra
            'image_mirrors': {
                'type': 'array',
                'description': 'Container image mirrors',
                'items': {
                    'type': 'object',
                    'properties': {
                        'from': {
                            'oneOf': [
                                {
                                    'type': 'string',
                                    'description': 'The source image',
                                    'pattern': '.+',
                                },
                                {
                                    'type': 'object',
                                    'properties': {
                                        'base': {
                                            'type': 'string',
                                        },
                                        'repositories': {
                                            'type': 'array',
                                            'items': {
                                                'type': 'string',
                                            },
                                        },
                                        'tags': {
                                            'type': 'array',
                                            'items': {
                                                'type': 'string',
                                            },
                                        },
                                    },
                                    'required': ['repositories'],
                                    'additionalProperties': False,
                                },
                            ],
                        },
                        'to': {
                            'oneOf': [
                                {
                                    'type': 'string',
                                    'description': 'The destination image',
                                },
                                {
                                    'type': 'object',
                                    'properties': {
                                        'base': {
                                            'type': 'string',
                                        },
                                        'repository': {
                                            'type': 'string',
                                        },
                                        'tag': {
                                            'type': 'string',
                                        },
                                    },
                                    'additionalProperties': False,
                                },
                            ],
                        },
                        'semver': {
                            'type': 'string',
                            'description': 'Version specification as an NPM range',
                        },
                    },
                    'required': ['from'],
                    'additionalProperties': False,
                },
                '_example': ('\n'
                             '    # Mirror debian:buster\n'
                             '    # ... to registry.example.org/mygroup/myproject/library/debian:buster:\n'
                             "    - from: 'debian:buster'\n"
                             '    # Overriding destination:\n'
                             "    - from: 'quay.org/coreos/etcd:v3.4.1'\n"
                             "      to: 'etcd:v3.4.1' # Default would be coreos/etcd:v3.4.1\n"),
                '_doc_link': 'image_mirrors.md',
            },
            # From https://docs.gitlab.com/ee/api/projects.html#edit-project
            'issues_enabled': {
                'type': 'boolean',
                'description': 'Enable issues for this project',
            },
            'merge_requests_enabled': {
                'type': 'boolean',
                'description': 'Enable merge requests for this project',
            },
            'jobs_enabled': {
                'type': 'boolean',
                'description': 'Enable jobs for this project',
            },
            'wiki_enabled': {
                'type': 'boolean',
                'description': 'Enable wiki for this project',
            },
            'snippets_enabled': {
                'type': 'boolean',
                'description': 'Enable snippets for this project',
            },
            # Below are undocumented settings
            'repository_storage': {
                'type': 'string',
                'description': 'Which storage shard the repository is on. Available only to admins',
            },
            'external_authorization_classification_label': {
                'type': 'string',
                'description': 'The classification label for the project',
            },
        },
        'additionalProperties': False,
    }

    IGNORED_PARAMS = ['initialize_with_readme',
                      'unknown_boards',
                      'unknown_board_lists',
                      'unknown_groups',
                      'unknown_labels',
                      'unknown_members',
                      'unknown_milestones',
                      'unknown_pipeline_schedules',
                      'unknown_pipeline_schedule_variables',
                      'unknown_protected_branches',
                      'unknown_protected_tags',
                      'unknown_variables',
                      'unknown_webhooks']

    CREATE_KEY = 'name'
    CREATE_PARAMS = ['initialize_with_readme']

    """"web_url()

    Returns the project's web URL.
    (Allows to mock the web URL).
    """
    def web_url(self):
        return self._obj.web_url

    def _get_current_branches(self):
        if not hasattr(self, '_current_branches'):
            try:
                self._current_branches = [branch.name for branch in self._obj.branches.list(all=True)]
            except GitlabListError as err:
                if err.response_code != 403:  # repository_enabled=false?
                    pass
                self._current_branches = None
        return self._current_branches

    """"_process_archived()

    Process the archived param.
    """
    def _process_archived(self, param_name, param_value, dry_run=False, skip_save=False):
        assert param_name == 'archived'  # noqa: S101
        assert not skip_save  # noqa: S101

        current_value = getattr(self._obj, param_name)
        if current_value != param_value:
            if dry_run:
                logger.info('[%s] NOT Changing param %s: %s -> %s (dry-run)',
                            self._name, param_name, current_value, param_value)
                setattr(self._obj, param_name, param_value)
            else:
                logger.info('[%s] Changing param %s: %s -> %s',
                            self._name, param_name, current_value, param_value)
                if param_value:
                    self._obj.archive()
                else:
                    self._obj.unarchive()

    """"_process_branches()

    Process the branches param.
    """
    def _process_branches(self, param_name, param_value, dry_run=False, skip_save=False):
        assert param_name == 'branches'  # noqa: S101
        assert not skip_save  # noqa: S101
        if 'default_branch' in self._content and self._content['default_branch'] in self._get_current_branches():
            # Create from target default branch if it exists
            ref = self._content['default_branch']
        elif self._obj.default_branch in self._get_current_branches():
            # Create from current default branch otherwise
            ref = self._obj.default_branch
        else:
            ref = None
        for branch_name in param_value:
            if branch_name not in self._get_current_branches():
                if ref is None:
                    logger.info('[%s] NOT Creating branch: %s (no reference)',
                                self._name, branch_name)
                elif dry_run:
                    logger.info('[%s] NOT Creating branch: %s (dry-run)',
                                self._name, branch_name)
                    self._current_branches.append(branch_name)
                else:
                    logger.info('[%s] Creating branch: %s',
                                self._name, branch_name)
                    self._obj.branches.create({
                        'branch': branch_name,
                        'ref': ref,
                    })
                    self._current_branches.append(branch_name)
            if branch_name in self._get_current_branches():
                # Next branch will be created from this ref
                ref = branch_name

    """"_process_protected_branches()

    Process the protected_branches param.
    """
    def _process_protected_branches(self, param_name, param_value, dry_run=False, skip_save=False):
        assert param_name == 'protected_branches'  # noqa: S101
        assert not skip_save  # noqa: S101
        unknown_protected_branches = self._content.get('unknown_protected_branches', 'warn')
        current_protected_branches = dict([[protected_branch.name, protected_branch]
                                          for protected_branch in self._obj.protectedbranches.list(all=True)])
        # We first check for already protected branches
        for protected_name, target_config in sorted(param_value.items()):
            target_config_int = {}
            for k, v in target_config.items():
                if k.endswith('_access_level'):
                    target_config_int[k] = access_level(v)
                else:
                    target_config_int[k] = v
            target_config_int['name'] = protected_name
            if protected_name in current_protected_branches:
                current_protected_branch = current_protected_branches[protected_name]
                current_config = {
                    'name': protected_name,
                    'push_access_level': current_protected_branch.push_access_levels[0]['access_level'],
                    'merge_access_level': current_protected_branch.merge_access_levels[0]['access_level'],
                }
            else:
                current_config = {}
            if current_config != target_config_int:
                if dry_run:
                    logger.info('[%s] NOT Changing protected branch %s access level: %s -> %s (dry-run)',
                                self._name, protected_name, current_config, target_config_int)
                else:
                    logger.info('[%s] Changing protected branch %s access level: %s -> %s',
                                self._name, protected_name, current_config, target_config_int)
                    if 'name' in current_config:
                        self._obj.protectedbranches.delete(protected_name)
                    try:
                        self._obj.protectedbranches.create(target_config_int)
                    except GitlabCreateError as err:
                        if err.response_code != 409:
                            pass
                        logger.warning('[%s] Unable to create protected branch %s: %s',
                                       self._name, protected_name, err.error_message)
        # Remaining protected branches
        if unknown_protected_branches not in ['ignore', 'skip']:
            for protected_name in current_protected_branches:
                if protected_name not in param_value:
                    if unknown_protected_branches in ['delete', 'remove']:
                        if dry_run:
                            logger.info('[%s] NOT Deleting unknown protected branch: %s (dry-run)',
                                        self._name, protected_name)
                        else:
                            logger.info('[%s] Deleting unknown protected branch: %s',
                                        self._name, protected_name)
                            self._obj.protectedbranches.delete(protected_name)
                    else:
                        logger.warning('[%s] NOT Deleting unknown protected branch: %s'
                                       ' (unknown_protected_branches=%s)',
                                       self._name, protected_name, unknown_protected_branches)

    """"_process_protected_tags()

    Process the protected_tags param.
    """
    def _process_protected_tags(self, param_name, param_value, dry_run=False, skip_save=False):
        assert param_name == 'protected_tags'  # noqa: S101
        assert not skip_save  # noqa: S101
        unknown_protected_tags = self._content.get('unknown_protected_tags', 'warn')
        try:
            current_protected_tags = dict([[protected_tag.name, protected_tag]
                                           for protected_tag in self._obj.protectedtags.list(all=True)])
        except AttributeError:
            logger.error('[%s] Unable to manage protected tags: %s',
                         self._name, 'protected tags requires python-gitlab >= 1.7.0')
            return
        # We first check for already protected tags
        for protected_name, target_config in sorted(param_value.items()):
            target_config = {
                'name': protected_name,
                'create_access_level': access_level(target_config),
            }
            if protected_name in current_protected_tags:
                current_protected_tag = current_protected_tags[protected_name]
                current_config = {
                    'name': protected_name,
                    'create_access_level': current_protected_tag.create_access_levels[0]['access_level'],
                }
            else:
                current_config = {}
            if current_config != target_config:
                if dry_run:
                    logger.info('[%s] NOT Changing protected tag %s access level: %s -> %s (dry-run)',
                                self._name, protected_name, current_config, target_config)
                else:
                    logger.info('[%s] Changing protected tag %s access level: %s -> %s',
                                self._name, protected_name, current_config, target_config)
                    if 'name' in current_config:
                        self._obj.protectedtags.delete(protected_name)
                    self._obj.protectedtags.create(target_config)
        # Remaining protected tags
        if unknown_protected_tags not in ['ignore', 'skip']:
            current_protected_tags = sorted(protected_tag.name
                                            for protected_tag in self._obj.protectedtags.list(all=True))
            for protected_name in current_protected_tags:
                if protected_name not in param_value:
                    if unknown_protected_tags in ['delete', 'remove']:
                        if dry_run:
                            logger.info('[%s] NOT Deleting unknown protected tag: %s (dry-run)',
                                        self._name, protected_name)
                        else:
                            logger.info('[%s] Deleting unknown protected tag: %s',
                                        self._name, protected_name)
                            self._obj.protectedtags.delete(protected_name)
                    else:
                        logger.warning('[%s] NOT Deleting unknown protected tag: %s (unknown_protected_tags=%s)',
                                       self._name, protected_name, unknown_protected_tags)

    """"_process_container_expiration_policy()

    Process the container_expiration_policy param.
    """
    def _process_container_expiration_policy(self, param_name, param_value, dry_run=False, skip_save=False):
        assert param_name == 'container_expiration_policy'  # noqa: S101
        assert not skip_save  # noqa: S101

        current_value = getattr(self._obj, param_name)

        for k, v in sorted(param_value.items()):
            current_v = current_value.get(k, None)
            if v != current_v:
                if dry_run:
                    logger.info('[%s] NOT Changing container expiration policy %s: %s -> %s (dry-run)',
                                self._name, k, current_v, v)
                else:
                    logger.info('[%s] Changing container expiration policy %s: %s -> %s',
                                self._name, k, current_v, v)
                    self._obj.container_expiration_policy_attributes = {k: v}
                    self._obj.save()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2013-2017 Gauvain Pocentek <gauvain@pocentek.net>
# Copyright (C) 2019-2020 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function

import argparse
import logging
import sys

import gitlabracadabra
import gitlabracadabra.parser
from gitlabracadabra.gitlab.connections import GitlabConnections


logger = logging.getLogger(__name__)


def _get_parser():
    parser = argparse.ArgumentParser(
        description='GitLabracadabra')
    parser.add_argument('--version', help='Display the version.',
                        action='store_true')
    parser.add_argument('-v', '--verbose', '--fancy',
                        help='Verbose mode',
                        action='store_true')
    parser.add_argument('-d', '--debug',
                        help='Debug mode (display HTTP requests)',
                        action='store_true')
    parser.add_argument('--logging-format',
                        help='Logging format',
                        choices=['short', 'long'],
                        default='short')
    parser.add_argument('-c', '--config-file', action='append',
                        help=('Configuration file to use. Can be used '
                              'multiple times.'))
    parser.add_argument('-g', '--gitlab',
                        help=('Which configuration section should '
                              'be used. If not defined, the default selection '
                              'will be used.'),
                        required=False)
    parser.add_argument('--dry-run',
                        help='Dry run',
                        action='store_true')
    parser.add_argument('--doc-markdown',
                        help=('Output the help for the given type (project, '
                              'group, user, application_settings) as '
                              'Markdown.'))
    parser.add_argument('action_files',
                        help='Action file. Can be used multiple times.',
                        metavar='ACTIONFILE.yml',
                        nargs='*',
                        default=['gitlabracadabra.yml'])

    return parser


def main():
    parser = _get_parser()

    args = parser.parse_args(sys.argv[1:])

    if args.version:
        print(gitlabracadabra.__version__)
        sys.exit(0)

    config_files = args.config_file
    gitlab_id = args.gitlab
    verbose = args.verbose
    debug = args.debug

    log_level = logging.WARNING
    if verbose:
        log_level = logging.INFO
    if debug:
        log_level = logging.DEBUG
    if args.logging_format == 'long':
        logging_format = '%(asctime)s [%(process)d] %(levelname)-8.8s %(name)s: %(message)s'
    else:
        logging_format = '%(levelname)-8.8s %(message)s'
    logging.basicConfig(format=logging_format, level=log_level)

    if args.doc_markdown:
        cls = gitlabracadabra.parser.GitlabracadabraParser.get_class_for(args.doc_markdown)
        print(cls.doc_markdown())
        sys.exit(0)

    try:
        GitlabConnections().load(gitlab_id, config_files, debug)
    except Exception as e:  # noqa: B902
        logger.error(str(e))  # noqa: G200
        sys.exit(1)

    # First pass: Load data and preflight checks
    objects = {}
    has_errors = False
    for action_file in args.action_files:
        if action_file.endswith('.yml') or action_file.endswith('.yaml'):
            parser = gitlabracadabra.parser.GitlabracadabraParser.from_yaml_file(action_file)
        else:
            logger.error('Unhandled file: %s', action_file)
            has_errors = True
            continue
        logger.debug('Parsing file: %s', action_file)
        objects[action_file] = parser.objects()
        for k, v in sorted(objects[action_file].items()):
            if len(v.errors()) > 0:
                for error in v.errors():
                    logger.error('Error in %s::: %s', k, error.message)
                has_errors = True

    if has_errors:
        logger.error('Preflight checks errors. Exiting')
        sys.exit(1)

    # Second pass:
    for action_file in args.action_files:
        for _name, obj in sorted(objects[action_file].items()):
            obj.process(args.dry_run)


if __name__ == '__main__':
    main()  # type: ignore
